"""bigquery-etl CLI monitoring command."""

import ast
import json
import os
import re
import sys
from collections import defaultdict
from pathlib import Path
from typing import Optional

import click
import sqlglot
from bigeye_sdk.authentication.api_authentication import APIKeyAuth
from bigeye_sdk.bigconfig_validation.validation_context import _BIGEYE_YAML_FILE_IX
from bigeye_sdk.client.datawatch_client import datawatch_client_factory
from bigeye_sdk.client.enum import Method
from bigeye_sdk.controller.metric_suite_controller import MetricSuiteController
from bigeye_sdk.exceptions.exceptions import FileLoadException
from bigeye_sdk.generated.com.bigeye.models.generated import MetricRunStatus
from bigeye_sdk.model.big_config import (
    BigConfig,
    ColumnSelector,
    RowCreationTimes,
    TagDeployment,
    TagDeploymentSuite,
)
from bigeye_sdk.model.protobuf_message_facade import (
    SimpleCollection,
    SimpleConstantThreshold,
    SimpleMetricDefinition,
    SimpleMetricSchedule,
    SimpleNamedSchedule,
    SimplePredefinedMetric,
    SimplePredefinedMetricName,
)

from bigquery_etl.config import ConfigLoader
from bigquery_etl.metadata.parse_metadata import METADATA_FILE, Metadata

from ..cli.utils import paths_matching_name_pattern, project_id_option, sql_dir_option
from ..util import extract_from_query_path
from ..util.common import render as render_template

BIGCONFIG_FILE = "bigconfig.yml"
CUSTOM_RULES_FILE = "bigeye_custom_rules.sql"
QUERY_FILE = "query.sql"
VIEW_FILE = "view.sql"
METRIC_STATUS_FAILURES = [
    MetricRunStatus.METRIC_RUN_STATUS_UPPERBOUND_CRITICAL,
    MetricRunStatus.METRIC_RUN_STATUS_LOWERBOUND_CRITICAL,
    MetricRunStatus.METRIC_RUN_STATUS_GROUPS_CRITICAL,
    MetricRunStatus.METRIC_RUN_STATUS_MUTABLE_UPPERBOUND_CRITICAL,
    MetricRunStatus.METRIC_RUN_STATUS_MUTABLE_LOWERBOUND_CRITICAL,
    MetricRunStatus.METRIC_RUN_STATUS_GROUPS_LIMIT_FAILED,
]


@click.group(
    help="""
        Commands for managing monitoring of datasets.
        """
)
@click.pass_context
def monitoring(ctx):
    """Create the CLI group for the monitoring command."""
    pass


@monitoring.command(
    help="""
    Deploy monitors defined in the BigConfig files to Bigeye.

    Requires BigConfig API key to be set via BIGEYE_API_KEY env variable.
    """
)
@click.argument("name")
@project_id_option("moz-fx-data-shared-prod")
@sql_dir_option
@click.option(
    "--workspace",
    default=463,
    help="Bigeye workspace to use when authenticating to API.",
)
@click.option(
    "--base-url",
    "--base_url",
    default="https://app.bigeye.com",
    help="Bigeye base URL.",
)
@click.option(
    "--dry-run",
    "--dry_run",
    "--dryrun",
    is_flag=True,
    help="Dry run the BigConfig files without applying them.",
    default=False,
)
@click.pass_context
def deploy(
    ctx,
    name: str,
    sql_dir: Optional[str],
    project_id: Optional[str],
    workspace: str,
    base_url: str,
    dry_run: bool,
) -> None:
    """Deploy Bigeye config."""
    api_key = os.environ.get("BIGEYE_API_KEY")
    if api_key is None:
        click.echo(
            "Bigeye API token needs to be set via `BIGEYE_API_KEY` env variable."
        )
        sys.exit(1)

    """Deploy monitors to Bigeye."""
    metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    if len(metadata_files) == 0:
        print(f"No metadata files matching {name}")
        return

    for metadata_file in list(set(metadata_files)):
        project, dataset, table = extract_from_query_path(metadata_file)
        try:
            metadata = Metadata.from_file(metadata_file)
            if metadata.monitoring and metadata.monitoring.enabled:
                ctx.invoke(
                    update,
                    name=metadata_file.parent,
                    sql_dir=sql_dir,
                    project_id=project_id,
                )
                api_auth = APIKeyAuth(base_url=base_url, api_key=api_key)
                client = datawatch_client_factory(api_auth, workspace_id=workspace)
                mc = MetricSuiteController(client=client)

                ctx.invoke(
                    validate,
                    name=metadata_file.parent,
                    sql_dir=sql_dir,
                    project_id=project_id,
                )

                if (
                    metadata_file.parent / VIEW_FILE
                ).exists() or "public_bigquery" in metadata.labels:
                    # monitoring to be deployed on a view
                    # Bigeye requires to explicitly set the partition column for views
                    if metadata.monitoring.partition_column and not dry_run:
                        ctx.invoke(
                            set_partition_column,
                            name=metadata_file.parent,
                            sql_dir=sql_dir,
                            project_id=project_id,
                        )

                if (metadata_file.parent / CUSTOM_RULES_FILE).exists():
                    if dry_run:
                        print(
                            f"Dry run; Skip deploying custom SQL rule {metadata_file.parent}"
                        )
                    else:
                        ctx.invoke(
                            deploy_custom_rules,
                            name=metadata_file.parent,
                            sql_dir=sql_dir,
                            project_id=project_id,
                        )

        except FileNotFoundError:
            print("No metadata file for: {}.{}.{}".format(project, dataset, table))

    try:
        # Deploy BigConfig files at once.
        # Deploying BigConfig files separately can lead to previously deployed metrics being removed.
        mc.execute_bigconfig(
            input_path=[
                metadata_file.parent / BIGCONFIG_FILE
                for metadata_file in list(set(metadata_files))
            ]
            + [f"{sql_dir}/{BIGCONFIG_FILE}"],
            output_path=Path(sql_dir).parent if sql_dir else None,
            apply=not dry_run,
            recursive=False,
            strict_mode=True,
            auto_approve=True,
        )
    except Exception as e:
        if dry_run:
            bigconfig_result = Path(f"{project_id}_PLAN.yml")
        else:
            bigconfig_result = Path(f"{project_id}_APPLY.yml")
        click.echo(bigconfig_result.read_text())
        raise e


def _sql_rules_from_file(custom_rules_file, project, dataset, table) -> list:
    """Extract the SQL rules from the custom rules file."""
    jinja_params = {
        **{
            "dataset_id": dataset,
            "table_name": table,
            "project_id": project,
            "format": False,
        },
    }

    rendered_result = render_template(
        custom_rules_file.name,
        template_folder=str(custom_rules_file.parent),
        templates_dir="",
        **jinja_params,
    )

    statements = []
    for statement in sqlglot.parse(rendered_result, read="bigquery"):
        if statement is None:
            continue

        for select_statement in statement.find_all(sqlglot.exp.Select):
            statements.append(select_statement)

    return statements


@monitoring.command(
    help="""
    Deploy custom SQL rules.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
@click.option(
    "--workspace",
    default=463,
    help="Bigeye workspace to use when authenticating to API.",
)
@click.option(
    "--base-url",
    "--base_url",
    default="https://app.bigeye.com",
    help="Bigeye base URL.",
)
def deploy_custom_rules(
    name: str,
    sql_dir: Optional[str],
    project_id: Optional[str],
    base_url: str,
    workspace: int,
) -> None:
    """Deploy custom SQL rules for files."""
    api_key = os.environ.get("BIGEYE_API_KEY")
    if api_key is None:
        click.echo(
            "Bigeye API token needs to be set via `BIGEYE_API_KEY` env variable."
        )
        sys.exit(1)

    custom_rules_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=[CUSTOM_RULES_FILE]
    )

    api_auth = APIKeyAuth(base_url=base_url, api_key=api_key)
    client = datawatch_client_factory(api_auth, workspace_id=workspace)
    collections = client.get_collections()
    warehouse_id = ConfigLoader.get("monitoring", "bigeye_warehouse_id")
    existing_rules = client.get_rules_for_source(warehouse_id=warehouse_id)
    existing_rules_sql = [rule.custom_rule.sql for rule in existing_rules.custom_rules]
    existing_schedules = {
        schedule.name: schedule.id
        for schedule in client.get_named_schedule().named_schedules
    }
    url = "/api/v1/custom-rules"

    for custom_rule_file in list(set(custom_rules_files)):
        project, dataset, table = extract_from_query_path(custom_rule_file)
        try:
            metadata = Metadata.from_file(custom_rule_file.parent / METADATA_FILE)
            if metadata.monitoring and metadata.monitoring.enabled:
                # Convert all the Airflow params to jinja usable dict.
                for select_statement in _sql_rules_from_file(
                    custom_rule_file, project, dataset, table
                ):
                    sql = select_statement.sql(dialect="bigquery")
                    if sql in existing_rules_sql:
                        continue

                    # parse config values from SQL comment
                    config = {}
                    if select_statement.comments:
                        config = ast.literal_eval("".join(select_statement.comments))

                    payload = {
                        "name": config.get(
                            "name", f"{project}_{dataset}_{table}_bqetl_check"
                        ),
                        "sql": sql,
                        "warehouseId": warehouse_id,
                        "thresholdType": "CUSTOM_RULES_THRESHOLD_TYPE_"
                        + config.get("alert_type", "count").upper(),
                    }

                    if "range" in config:
                        if "min" in config["range"]:
                            payload["lowerThreshold"] = config["range"]["min"]
                        if "max" in config["range"]:
                            payload["upperThreshold"] = config["range"]["max"]

                    if "owner" in config:
                        payload["owner"] = {"email": config["owner"]}

                    if "collections" in config:
                        collection_ids = [
                            collection.id
                            for collection in collections.collections
                            if collection.name in config["collections"]
                        ]
                        payload["collectionIds"] = collection_ids

                    if (
                        "schedule" in config
                        and config["schedule"] in existing_schedules
                    ):
                        payload["metricSchedule"] = {
                            "namedSchedule": {
                                "id": existing_schedules[config["schedule"]]
                            }
                        }

                    try:
                        response = client._call_datawatch(
                            Method.POST,
                            url=url,
                            body=json.dumps({"customRule": payload}),
                        )

                        if "id" in response:
                            click.echo(
                                f"Created custom rule {response['id']} for `{project}.{dataset}.{table}`"
                            )
                    except Exception as e:
                        if "There was an error processing your request" in str(e):
                            # API throws an error when partition column was already set.
                            # There is no API endpoint to check for partition columns though
                            click.echo(
                                f"Partition column for `{project}.{dataset}.{table}` already set."
                            )
                        else:
                            raise e

        except FileNotFoundError:
            print("No metadata file for: {}.{}.{}".format(project, dataset, table))


def _update_bigconfig(
    bigconfig,
    metadata,
    project,
    dataset,
    table,
    default_metrics,
):
    """Update the BigConfig file to monitor a view."""
    for collection in bigconfig.tag_deployments:
        for deployment in collection.deployments:
            for metric in deployment.metrics:
                if (
                    metric.metric_type is not None
                    and metric.metric_type.predefined_metric in default_metrics
                ):
                    default_metrics.remove(metric.metric_type.predefined_metric)
                elif (
                    metric.saved_metric_id is not None
                    and (metric_id := metric.saved_metric_id.upper()) in default_metrics
                ):
                    default_metrics.remove(metric_id)

        if metadata.monitoring.collection and collection.collection is None:
            collection.collection = SimpleCollection(
                name=metadata.monitoring.collection
            )

    if len(default_metrics) > 0:
        deployments = [
            TagDeployment(
                column_selectors=[
                    ColumnSelector(name=f"{project}.{project}.{dataset}.{table}.*")
                ],
                metrics=[
                    SimpleMetricDefinition(
                        metric_name=(
                            f"{metric.name}"
                            if "volume" not in metric.name.lower()
                            else f"{metric.name} [fail]"
                        ),
                        metric_type=SimplePredefinedMetric(
                            type="PREDEFINED", predefined_metric=metric
                        ),
                        metric_schedule=SimpleMetricSchedule(
                            named_schedule=SimpleNamedSchedule(
                                name="Default Schedule - 13:00 UTC"
                            )
                        ),
                    )
                    for metric in default_metrics
                ],
            )
        ]

        collection = None
        if metadata.monitoring.collection:
            collection = SimpleCollection(name=metadata.monitoring.collection)

        bigconfig.tag_deployments += [
            TagDeploymentSuite(deployments=deployments, collection=collection)
        ]

        if metadata.monitoring.partition_column:
            bigconfig.row_creation_times = RowCreationTimes(
                column_selectors=[
                    ColumnSelector(
                        name=f"{project}.{project}.{dataset}.{table}.{metadata.monitoring.partition_column}"
                    )
                ]
            )


@monitoring.command(
    help="""
    Update BigConfig files based on monitoring metadata.

    Requires BigConfig credentials to be set via BIGEYE_API_CRED_FILE env variable.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
def update(name: str, sql_dir: Optional[str], project_id: Optional[str]) -> None:
    """Update BigConfig files based on monitoring metadata."""
    metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    for metadata_file in list(set(metadata_files)):
        project, dataset, table = extract_from_query_path(metadata_file)
        try:
            metadata = Metadata.from_file(metadata_file)
            if metadata.monitoring and metadata.monitoring.enabled:
                bigconfig_file = metadata_file.parent / BIGCONFIG_FILE

                if bigconfig_file.exists():
                    bigconfig = BigConfig.load(bigconfig_file)
                else:
                    bigconfig = BigConfig(type="BIGCONFIG_FILE")

                if (
                    metadata_file.parent / QUERY_FILE
                ).exists() and "public_bigquery" not in metadata.labels:
                    _update_bigconfig(
                        bigconfig=bigconfig,
                        metadata=metadata,
                        project=project,
                        dataset=dataset,
                        table=table,
                        default_metrics=[
                            SimplePredefinedMetricName.FRESHNESS,
                            SimplePredefinedMetricName.VOLUME,
                        ],
                    )

                bigconfig.save(
                    output_path=bigconfig_file.parent,
                    default_file_name=bigconfig_file.stem,
                )

        except FileNotFoundError:
            print("No metadata file for: {}.{}.{}".format(project, dataset, table))


@monitoring.command(
    help="""
    Validate BigConfig files.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
def validate(name: str, sql_dir: Optional[str], project_id: Optional[str]) -> None:
    """Validate BigConfig file."""
    bigconfig_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["bigconfig.yml"]
    )

    invalid = False

    for bigconfig_file in list(set(bigconfig_files)):
        try:
            BigConfig.load(bigconfig_file)
        except FileLoadException as e:
            if "Duplicate" in e.message:
                pass
            else:
                click.echo(f"Invalid BigConfig file {bigconfig_file}: {e}")
                invalid = True
        except Exception as e:
            click.echo(f"Invalid BigConfig file {bigconfig_file}: {e}")
            invalid = True

        if (bigconfig_file.parent / VIEW_FILE).exists():
            try:
                metadata = Metadata.from_file(bigconfig_file.parent / METADATA_FILE)
                if metadata.monitoring and metadata.monitoring.enabled:
                    if not metadata.monitoring.partition_column_set:
                        invalid = True
                        click.echo(
                            "If montoring is enabled for views, then the partition_column needs to be configured "
                            + f"explicitly in the `monitoring` metadata for {bigconfig_file.parent}. "
                            + "Set `partition_column: null` for views that do not reference a partitioned table."
                        )
            except FileNotFoundError:
                pass  # view has no metadata file, so monitoring is disabled

    if invalid:
        sys.exit(1)

    click.echo("All BigConfig files are valid.")


@monitoring.command(
    help="""
    Set partition column for view or table in Bigeye.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
@click.option(
    "--workspace",
    default=463,
    help="Bigeye workspace to use when authenticating to API.",
)
@click.option(
    "--base-url",
    "--base_url",
    default="https://app.bigeye.com",
    help="Bigeye base URL.",
)
def set_partition_column(
    name: str,
    sql_dir: Optional[str],
    project_id: Optional[str],
    base_url: str,
    workspace: int,
) -> None:
    """Validate BigConfig file."""
    api_key = os.environ.get("BIGEYE_API_KEY")
    if api_key is None:
        click.echo(
            "Bigeye API token needs to be set via `BIGEYE_API_KEY` env variable."
        )
        sys.exit(1)

    metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    for metadata_file in list(set(metadata_files)):
        project, dataset, table = extract_from_query_path(metadata_file)
        try:
            metadata = Metadata.from_file(metadata_file)
            if metadata.monitoring and metadata.monitoring.enabled:
                if not metadata.monitoring.partition_column_set:
                    click.echo(f"No partition column set for {metadata_file.parent}")
                    continue
                api_auth = APIKeyAuth(base_url=base_url, api_key=api_key)
                client = datawatch_client_factory(api_auth, workspace_id=workspace)

                warehouse_id = ConfigLoader.get("monitoring", "bigeye_warehouse_id")
                table_id = client.get_table_ids(
                    warehouse_id=warehouse_id,
                    schemas=f"{project}.{dataset}",
                    table_name=table,
                )

                if len(table_id) == 0:
                    click.echo(
                        f"Table `{project}.{dataset}.{table}` does not exist in Bigeye."
                    )
                    sys.exit(1)
                table_id = table_id[0]

                columns = client.get_columns(table_ids=[table_id])
                column_id = None
                for column in columns.columns:
                    if column.column.name == metadata.monitoring.partition_column:
                        column_id = column.column.id
                        break

                if column_id is None:
                    click.echo(
                        f"Configured partition column {metadata.monitoring.partition_column} does not exist for table `{project}.{dataset}.{table}`"
                    )
                    sys.exit(1)

                url = f"/api/v1/tables/{table_id}/required-partition-column"
                payload = json.dumps({"columnId": column_id})

                try:
                    response = client._call_datawatch(Method.PUT, url=url, body=payload)

                    if (
                        response.get("datasetId", None) == table_id
                        and response.get("requiredPartitionColumnId") == column_id
                    ):
                        click.echo(
                            f"Set partition column {column_id} for `{project}.{dataset}.{table}`"
                        )
                except Exception as e:
                    if "There was an error processing your request" in str(
                        e
                    ) or "Dataset already requires a partition filter" in str(e):
                        # API throws an error when partition column was already set.
                        # There is no API endpoint to check for partition columns though
                        pass
                    else:
                        raise e

        except FileNotFoundError:
            print("No metadata file for: {}.{}.{}".format(project, dataset, table))


@monitoring.command(
    help="""
    Delete deployed monitors. Use --custom-sql and/or --metrics flags to select which types of monitors to delete.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
@click.option(
    "--workspace",
    default=463,
    help="Bigeye workspace to use when authenticating to API.",
)
@click.option(
    "--base-url",
    "--base_url",
    default="https://app.bigeye.com",
    help="Bigeye base URL.",
)
@click.option(
    "--custom-sql",
    "--custom_sql",
    is_flag=True,
    default=False,
    help="Deletes custom SQL rules.",
)
@click.option(
    "--metrics",
    is_flag=True,
    default=False,
    help="Deletes metrics checks.",
)
def delete(
    name: str,
    sql_dir: Optional[str],
    project_id: Optional[str],
    base_url: str,
    workspace: int,
    custom_sql: bool,
    metrics: bool,
) -> None:
    """Validate BigConfig file."""
    api_key = os.environ.get("BIGEYE_API_KEY")
    if api_key is None:
        click.echo(
            "Bigeye API token needs to be set via `BIGEYE_API_KEY` env variable."
        )
        sys.exit(1)

    metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )
    api_auth = APIKeyAuth(base_url=base_url, api_key=api_key)
    client = datawatch_client_factory(api_auth, workspace_id=workspace)
    warehouse_id = ConfigLoader.get("monitoring", "bigeye_warehouse_id")
    existing_rules = {
        rule.custom_rule.sql: rule.id
        for rule in client.get_rules_for_source(warehouse_id=warehouse_id).custom_rules
    }

    for metadata_file in list(set(metadata_files)):
        project, dataset, table = extract_from_query_path(metadata_file)
        if metrics:
            deployed_metrics = client.get_metric_info_batch_post(
                table_name=table,
                schema_name=f"{project}.{dataset}",
                warehouse_ids=[warehouse_id],
            )
            client.delete_metrics(metrics=deployed_metrics.metrics)

        if custom_sql:
            if (metadata_file.parent / CUSTOM_RULES_FILE).exists():
                for select_statement in _sql_rules_from_file(
                    metadata_file.parent / CUSTOM_RULES_FILE, project, dataset, table
                ):
                    sql = select_statement.sql(dialect="bigquery")
                    if sql in existing_rules:
                        client.delete_custom_rule(existing_rules[sql])
                        click.echo(
                            f"Deleted custom rule {existing_rules[sql]} for {project}.{dataset}.{table}"
                        )


@monitoring.command(
    help="""
    Runs Bigeye monitors.

    Example:

    \t./bqetl monitoring run ga_derived.downloads_with_attribution_v2
    """,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("name")
@project_id_option()
@sql_dir_option
@click.option(
    "--workspace",
    default=463,
    help="Bigeye workspace to use when authenticating to API.",
)
@click.option(
    "--base-url",
    "--base_url",
    default="https://app.bigeye.com",
    help="Bigeye base URL.",
)
@click.option("--marker", default="", help="Marker to filter checks.")
def run(name, project_id, sql_dir, workspace, base_url, marker):
    """Run Bigeye checks."""
    api_key = os.environ.get("BIGEYE_API_KEY")
    if api_key is None:
        click.echo(
            "Bigeye API token needs to be set via `BIGEYE_API_KEY` env variable."
        )
        sys.exit(1)

    api_auth = APIKeyAuth(base_url=base_url, api_key=api_key)
    client = datawatch_client_factory(api_auth, workspace_id=workspace)
    warehouse_id = ConfigLoader.get("monitoring", "bigeye_warehouse_id")
    existing_rules = {
        rule.custom_rule.sql: {"id": rule.id, "name": rule.custom_rule.name}
        for rule in client.get_rules_for_source(warehouse_id=warehouse_id).custom_rules
        if rule.custom_rule.name.endswith(marker or "")
    }

    metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    failed = False

    for metadata_file in list(set(metadata_files)):
        project, dataset, table = extract_from_query_path(metadata_file)
        try:
            metadata = Metadata.from_file(metadata_file)
            if metadata.monitoring and metadata.monitoring.enabled:
                metrics = client.get_metric_info_batch_post(
                    table_name=table,
                    schema_name=f"{project}.{dataset}",
                    warehouse_ids=[warehouse_id],
                ).metrics

                if marker:
                    metrics = [
                        metric for metric in metrics if metric.name.endswith(marker)
                    ]

                metric_ids = [metric.metric_configuration.id for metric in metrics]

                click.echo(
                    f"Trigger metric runs for {project}.{dataset}.{table}: {metric_ids}"
                )
                response = client.run_metric_batch_async(metric_ids=metric_ids)
                for metric_info in response:
                    latest_metric_run = metric_info.latest_metric_runs[-1]
                    if (
                        latest_metric_run
                        and latest_metric_run.status in METRIC_STATUS_FAILURES
                    ):
                        if metric_info.metric_configuration.name.lower().endswith(
                            "[fail]"
                        ):
                            failed = True
                        click.echo(
                            f"Error running check {metric_info.metric_configuration.id}: {metric_info.active_issue.display_name}"
                        )
                        click.echo(
                            f"Check {base_url}/w/{workspace}/catalog/data-sources/metric/{metric_info.metric_configuration.id}/chart for more information."
                        )

                if (metadata_file.parent / CUSTOM_RULES_FILE).exists():
                    for select_statement in _sql_rules_from_file(
                        metadata_file.parent / CUSTOM_RULES_FILE,
                        project,
                        dataset,
                        table,
                    ):
                        sql = select_statement.sql(dialect="bigquery")
                        if sql in existing_rules:
                            response = client._call_datawatch(
                                Method.GET,
                                url=f"/api/v1/custom-rules/run/{existing_rules[sql]['id']}",
                            )
                            click.echo(
                                f"Triggered custom rule {existing_rules[sql]['id']} for {project}.{dataset}.{table}"
                            )

                            latest_rule_run = response.get("latestRuns", [])
                            if latest_rule_run and latest_rule_run[-1].get(
                                "status"
                            ) in {status.name for status in METRIC_STATUS_FAILURES}:
                                if (
                                    not existing_rules[sql]["name"]
                                    .lower()
                                    .endswith("[warn]")
                                ):
                                    failed = True

                                click.echo(
                                    f"Error running custom rule {existing_rules[sql]} for {project}.{dataset}.{table}. "
                                    + f"Check {base_url}/w/{workspace}/catalog/data-sources/{warehouse_id}/rules/{existing_rules[sql]['id']}/runs "
                                    + "for more information."
                                )

        except FileNotFoundError:
            print("No metadata file for: {}.{}.{}".format(project, dataset, table))

    if failed:
        sys.exit(1)


# TODO: remove this command once checks have been migrated
@monitoring.command(
    help="""
    Create BigConfig files from ETL check.sql files.

    This is a temporary command and will be removed after checks have been migrated.
    """
)
@click.argument("name")
@project_id_option()
@sql_dir_option
@click.pass_context
def migrate(ctx, name: str, sql_dir: Optional[str], project_id: Optional[str]) -> None:
    """Migrate checks.sql files to BigConfig."""
    check_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["checks.sql"]
    )

    for check_file in list(set(check_files)):
        project, dataset, table = extract_from_query_path(check_file)
        checks_migrated = 0
        try:
            metadata_file = check_file.parent / METADATA_FILE
            metadata = Metadata.from_file(check_file.parent / METADATA_FILE)
            metadata.monitoring = {"enabled": True}
            metadata.write(metadata_file)

            ctx.invoke(
                update,
                name=metadata_file.parent,
                sql_dir=sql_dir,
                project_id=project_id,
            )

            # clearing Bigeye file index after running the `update` command
            # this is necessary as Bigeye throws an exception when it sees the same config file twice,
            # which will be the case after doing `update` previously
            _BIGEYE_YAML_FILE_IX.clear()
            bigconfig_file = check_file.parent / BIGCONFIG_FILE
            if bigconfig_file.exists():
                bigconfig = BigConfig.load(bigconfig_file)
            else:
                bigconfig = BigConfig(type="BIGCONFIG_FILE")

            checks = check_file.read_text()
            metrics_by_column = defaultdict(list)

            # map pre-defined ETL checks to Bigeye checks
            if matches := re.findall(
                r"not_null\(\[([^\]\)]*)\](?:, \"(.*)\")?\)", checks
            ):
                for match in matches:
                    checks_migrated += 1
                    columns = [
                        col.replace('"', "")
                        for col in match[0]
                        .replace("[", "")
                        .replace("]", "")
                        .split(", ")
                    ]
                    metric_definition = SimpleMetricDefinition(
                        metric_type=SimplePredefinedMetric(
                            type="PREDEFINED",
                            predefined_metric=SimplePredefinedMetricName.PERCENT_NOT_NULL,
                        ),
                        threshold=SimpleConstantThreshold(
                            type="CONSTANT", lower_bound=1.0
                        ),
                    )

                    for column in columns:
                        metrics_by_column[
                            f"{project}.{project}.{dataset}.{table}.{column}"
                        ].append(metric_definition)

            if matches := re.findall(
                r"min_row_count\(([^,\)]*)(?:, \"(.*)\")?\)", checks
            ):
                for match in matches:
                    checks_migrated += 1
                    metrics_by_column[
                        f"{project}.{project}.{dataset}.{table}.*"
                    ].append(
                        SimpleMetricDefinition(
                            metric_type=SimplePredefinedMetric(
                                type="PREDEFINED",
                                predefined_metric=SimplePredefinedMetricName.COUNT_ROWS,
                            ),
                            threshold=SimpleConstantThreshold(
                                type="CONSTANT", lower_bound=int(match[0])
                            ),
                        )
                    )

            if matches := re.findall(r"is_unique\(([^\]\)]*)(?:, \"(.*)\")?\)", checks):
                for match in matches:
                    checks_migrated += 1
                    columns = [
                        col.replace('"', "")
                        for col in match[0]
                        .replace("[", "")
                        .replace("]", "")
                        .split(", ")
                    ]
                    metric_definition = SimpleMetricDefinition(
                        metric_type=SimplePredefinedMetric(
                            type="PREDEFINED",
                            predefined_metric=SimplePredefinedMetricName.COUNT_DUPLICATES,
                        ),
                        threshold=SimpleConstantThreshold(
                            type="CONSTANT", lower_bound=0.0
                        ),
                    )

                    for column in columns:
                        metrics_by_column[
                            f"{project}.{project}.{dataset}.{table}.{column}"
                        ].append(metric_definition)

            if matches := re.findall(r"in_range\((.*), (.*), (.*), \"(.*)\"\)", checks):
                for match in matches:
                    checks_migrated += 1
                    columns = [
                        col.replace('"', "")
                        for col in match[0]
                        .replace("[", "")
                        .replace("]", "")
                        .split(", ")
                    ]
                    metric_definition = SimpleMetricDefinition(
                        metric_name="Range",
                        metric_type=SimplePredefinedMetric(
                            type="PREDEFINED",
                            predefined_metric=SimplePredefinedMetricName.MIN,
                        ),
                        threshold=SimpleConstantThreshold(
                            type="CONSTANT",
                            lower_bound=int(match[1]) if match[1] != "none" else None,
                            upper_bound=int(match[2]) if match[2] != "none" else None,
                        ),
                    )

                    for column in columns:
                        metrics_by_column[
                            f"{project}.{project}.{dataset}.{table}.{column}"
                        ].append(metric_definition)

            if matches := re.findall(
                r"value_length\(column=\"(.*)\", expected_length=(.*), where=\"(.*)\"\)",
                checks,
            ):
                for match in matches:
                    checks_migrated += 1
                    metric_definition = SimpleMetricDefinition(
                        metric_name="Value Length",
                        metric_type=SimplePredefinedMetric(
                            type="PREDEFINED",
                            predefined_metric=SimplePredefinedMetricName.STRING_LENGTH_MIN,
                        ),
                        threshold=SimpleConstantThreshold(
                            type="CONSTANT",
                            lower_bound=int(match[1]),
                            upper_bound=int(match[1]),
                        ),
                    )

                    for column in columns:
                        metrics_by_column[
                            f"{project}.{project}.{dataset}.{table}.{column}"
                        ].append(metric_definition)

            deployments = []
            for column_selector, metrics in metrics_by_column.items():
                deployments.append(
                    TagDeployment(
                        column_selectors=[ColumnSelector(name=column_selector)],
                        metrics=metrics,
                    )
                )

            bigconfig.tag_deployments += [TagDeploymentSuite(deployments=deployments)]

            bigconfig.save(
                output_path=bigconfig_file.parent,
                default_file_name=bigconfig_file.stem,
            )

            total_checks = checks.count("#fail") + checks.count("#warn")
            click.echo(
                f"Migrated {checks_migrated} of {total_checks} checks to {bigconfig_file}."
            )
            if checks_migrated < total_checks:
                click.echo(
                    f"There might be custom SQL checks that need to be migrated manually for {check_file.parent}"
                )

        except FileNotFoundError:
            print(f"No metadata file for: {check_file.parent}")

    click.echo(
        "Please manually check the migration logic as it might not be 100% correct"
    )



#!/usr/bin/python3

"""Read a table from BigQuery and write it as parquet."""

import json
import re
import sys
from argparse import ArgumentParser
from textwrap import dedent

try:
    from google.cloud import bigquery
except ImportError as e:
    bigquery = None  # type: ignore
    bigquery_error = e

parser = ArgumentParser(description=__doc__)
parser.add_argument("table", help="BigQuery table to read")
parser.add_argument(
    "--avro-path",
    help="Avro export path that should be read instead of BigQuery Storage API.",
)
parser.add_argument(
    "--dataset",
    dest="dataset",
    default="telemetry",
    help='BigQuery dataset that contains TABLE. Defaults to "telemetry".',
)
parser.add_argument(
    "--destination",
    default="s3://telemetry-parquet/",
    help="The path where parquet will be written. This will have table and"
    " --static-partitions appended as directories. If this starts with s3:// it will"
    " be replaced with s3a://.",
)
parser.add_argument(
    "--destination-table",
    default=None,
    help="The name of the destination parquet table. Defaults to the source table name."
    " If this ends with a version by matching /_v[0-9]+$/ it will be converted to"
    " a directory in the output.",
)
parser.add_argument(
    "--drop",
    default=[],
    dest="drop",
    nargs="+",
    help="A list of fields to exclude from the output. Passed to df.drop(*_) after"
    " --select is applied.",
)
parser.add_argument(
    "--dry-run",
    dest="dry_run",
    action="store_true",
    help="Print the spark job that would be run, without running it, and exit.",
)
parser.add_argument(
    "--filter",
    default=[],
    dest="filter",
    nargs="+",
    help="A list of conditions to limit the input before it reaches spark. Passed to"
    ' spark.read.format("bigquery").option("filter", " AND ".join(_)). If a field is'
    " referenced here and later dropped via --drop or --static-partitions then it must"
    " be referenced in --where.",
)
parser.add_argument(
    "--partition-by",
    default=["submission_date"],
    dest="partition_by",
    nargs="+",
    help="A list of fields on which to dynamically partition the output. Passed to"
    ' df.write.partitionBy(*_). Defaults to ["submission_date"]. Fields specified in'
    " --static-partitions will be ignored.",
)
parser.add_argument(
    "--select",
    default=["*"],
    dest="select",
    nargs="+",
    help="A list of all fields to include in the output."
    " Passed to df.selectExpr(*_) after --replace is applied and before --drop."
    " Can include sql expressions.",
)
parser.add_argument(
    "--maps-from-entries",
    action="store_true",
    help="Recursively convert repeated key-value structs with maps.",
)
parser.add_argument(
    "--bigint-columns",
    nargs="*",
    help="A list of columns that should remain BIGINT in the output, while any other "
    " BIGINT columns are converted to INT. If unspecified, all columns remain BIGINT.",
)
parser.add_argument(
    "--replace",
    default=[],
    nargs="+",
    help="A list of expressions that modify columns in the output. Passed to"
    " df.withColumnExpr(_) one at a time after --where applied and before --select."
    " Can include sql expressions.",
)
parser.add_argument(
    "--static-partitions",
    dest="static_partitions",
    default=[],
    nargs="+",
    help="Static partitions specified as FIELD=VALUE that will be appended to"
    " --destination after table, with FIELD dropped from the input.",
)
parser.add_argument(
    "--submission-date",
    dest="submission_date",
    help="Short for: --filter \"submission_date = DATE 'SUBMISSION_DATE'\""
    " --where \"submission_date = DATE 'SUBMISSION_DATE'\""
    " --static-partitions submission_date=SUBMISSION_DATE",
)
parser.add_argument(
    "--where",
    dest="where",
    default="TRUE",
    help="An expression to limit the output. Passed to df.where(_). If a field is"
    " referenced in filter and later dropped via --drop or --static-partitions then it"
    " must be referenced here.",
)
parser.add_argument(
    "--write-mode",
    dest="write_mode",
    default="overwrite",
    help='Passed to df.write.mode(_). Defaults to "overwrite".',
)
parser.add_argument(
    "--partition-overwrite-mode",
    default="STATIC",
    type=str.upper,
    help='Passed to spark.conf.set("spark.sql.sources.partitionOverwriteMode", _).'
    ' Defaults to "STATIC".',
)


def transform_field(
    field, maps_from_entries=False, bigint_columns=None, transform_layer=0, *prefix
):
    """
    Generate spark SQL to recursively convert fields types.

    If maps_from_entries is True, convert repeated key-value structs to maps.

    If bigint_columns is a list, convert non-matching BIGINT columns to INT.
    """
    transformed = False
    result = full_name = ".".join(prefix + (field.name,))
    repeated = field.mode == "REPEATED"
    if repeated:
        if transform_layer > 0:
            prefix = (f"_{transform_layer}",)
        else:
            prefix = ("_",)
        transform_layer += 1
    else:
        prefix = (*prefix, field.name)
    if field.field_type == "RECORD":
        if bigint_columns is not None:
            # get the bigint_columns nested under this field
            prefix_len = len(field.name) + 1
            bigint_columns = [
                column[prefix_len:]
                for column in bigint_columns
                if column.startswith(field.name + ".")
            ]
        subfields = [
            transform_field(
                subfield, maps_from_entries, bigint_columns, transform_layer, *prefix
            )
            for subfield in field.fields
        ]
        if any(subfield_transformed for _, subfield_transformed in subfields):
            transformed = True
            fields = ", ".join(transform for transform, _ in subfields)
            result = f"STRUCT({fields})"
            if repeated:
                result = f"TRANSFORM({full_name}, {prefix[0]} -> {result})"
        if maps_from_entries:
            if repeated and {"key", "value"} == {f.name for f in field.fields}:
                transformed = True
                result = f"MAP_FROM_ENTRIES({result})"
    elif field.field_type == "INTEGER":
        if bigint_columns is not None and field.name not in bigint_columns:
            transformed = True
            if repeated:
                result = f"TRANSFORM({full_name}, {prefix[0]} -> INT({prefix[0]}))"
            else:
                result = f"INT({full_name})"
    return f"{result} AS {field.name}", transformed


def transform_schema(table, maps_from_entries=False, bigint_columns=None):
    """Get maps_from_entries expressions for all maps in the given table."""
    if bigquery is None:
        return ["..."]
    schema = sorted(bigquery.Client().get_table(table).schema, key=lambda f: f.name)
    replace = []
    for index, field in enumerate(schema):
        try:
            expr, transformed = transform_field(
                field, maps_from_entries, bigint_columns
            )
        except Exception:
            json_field = json.dumps(field.to_api_repr(), indent=2)
            print(f"problem with field {index}:\n{json_field}", file=sys.stderr)
            raise
        if transformed:
            replace += [expr]
    return replace


def main():
    """Read a table from BigQuery and write it as parquet."""
    args = parser.parse_args()

    # handle --submission-date
    if args.submission_date is not None:
        # --filter "submission_date = DATE 'SUBMISSION_DATE'"
        condition = "submission_date = DATE '" + args.submission_date + "'"
        args.filter.append(condition)
        # --static-partitions submission_date=SUBMISSION_DATE
        args.static_partitions.append("submission_date=" + args.submission_date)
        # --where "submission_date IS NOT NULL"
        if args.where == "TRUE":
            args.where = condition
        else:
            args.where = "(" + args.where + ") AND " + condition

    # Set default --destination-table if it was not provided
    if args.destination_table is None:
        args.destination_table = args.table

    # append table and --static-partitions to destination
    args.destination = "/".join(
        [
            re.sub("^s3://", "s3a://", args.destination).rstrip("/"),
            re.sub("_(v[0-9]+)$", r"/\1", args.destination_table.rsplit(".", 1).pop()),
        ]
        + args.static_partitions
    )

    # convert --static-partitions to a dict
    args.static_partitions = dict(p.split("=", 1) for p in args.static_partitions)

    # remove --static-partitions fields from --partition-by
    args.partition_by = [
        f for f in args.partition_by if f not in args.static_partitions
    ]

    # add --static-partitions fields to --drop
    args.drop += args.static_partitions.keys()

    # convert --filter to a single string
    args.filter = " AND ".join(args.filter)

    if args.maps_from_entries or args.bigint_columns is not None:
        if "." in args.table:
            table_ref = args.table.replace(":", ".")
        else:
            table_ref = f"{args.dataset}.{args.table}"
        args.replace += transform_schema(
            table_ref, args.maps_from_entries, args.bigint_columns
        )

    if args.dry_run:
        replace = f"{args.replace!r}"
        if len(replace) > 60:
            replace = (
                "["
                + ",".join(f"\n{' '*4*5}{expr!r}" for expr in args.replace)
                + f"\n{' '*4*4}]"
            )
        print("spark = SparkSession.builder.appName('export_to_parquet').getOrCreate()")
        print("")
        print(
            "spark.conf.set('spark.sql.sources.partitionOverwriteMode', "
            f"{args.partition_overwrite_mode!r})"
        )
        print("")
        if args.avro_path is not None:
            print(f"df = spark.read.format('avro').load({args.avro_path!r})")
        else:
            print(
                dedent(
                    f"""
                    df = (
                        spark.read.format('bigquery')
                        .option('dataset', {args.dataset!r})
                        .option('table', {args.table!r})
                        .option('filter', {args.filter!r})
                        .option("parallelism", 0)  # let BigQuery storage API decide
                        .load()
                    )
                    """
                ).strip()
            )
        print("")
        print(
            dedent(
                f"""
                df = df.where({args.where!r}).selectExpr(*{args.select!r}).drop(*{args.drop!r})

                for sql in {replace}:
                    value, name = re.fullmatch("(?i)(.*) AS (.*)", sql).groups()
                    df = df.withColumn(name, expr(value))

                (
                    df.write.mode({args.write_mode!r})
                    .partitionBy(*{args.partition_by!r})
                    .parquet({args.destination!r})
                )
                """  # noqa:E501
            ).strip()
        )
    else:
        # delay import to allow --dry-run without spark
        from pyspark.sql import SparkSession
        from pyspark.sql.functions import expr

        if bigquery is None:
            raise bigquery_error

        spark = SparkSession.builder.appName("export_to_parquet").getOrCreate()

        spark.conf.set(
            "spark.sql.sources.partitionOverwriteMode", args.partition_overwrite_mode
        )

        # run spark job from parsed args
        if args.avro_path is not None:
            df = spark.read.format("avro").load(args.avro_path)
        else:
            df = (
                spark.read.format("bigquery")
                .option("dataset", args.dataset)
                .option("table", args.table)
                .option("filter", args.filter)
                .option("parallelism", 0)  # let BigQuery storage API decide
                .load()
            )

        df = df.where(args.where).selectExpr(*args.select).drop(*args.drop)

        for sql in args.replace:
            value, name = re.fullmatch("(?i)(.*) AS (.*)", sql).groups()
            df = df.withColumn(name, expr(value))

        (
            df.write.mode(args.write_mode)
            .partitionBy(*args.partition_by)
            .parquet(args.destination)
        )


if __name__ == "__main__":
    main()



"""Utilities for deploying BigQuery tables from bigquery-etl queries."""

import logging
from pathlib import Path
from typing import Optional

from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from .config import ConfigLoader
from .dryrun import DryRun, get_id_token
from .metadata.parse_metadata import METADATA_FILE, Metadata
from .metadata.publish_metadata import attach_external_data_config, attach_metadata
from .schema import SCHEMA_FILE, Schema

log = logging.getLogger(__name__)


class SkippedDeployException(Exception):
    """Raised when a deployment is skipped."""


class FailedDeployException(Exception):
    """Raised for failed deployments."""


class SkippedExternalDataException(Exception):
    """Raised for external data tables if skip_external_data is True."""


def deploy_table(
    artifact_file: Path,
    destination_table: Optional[str] = None,
    force: bool = False,
    use_cloud_function: bool = False,
    skip_existing: bool = False,
    skip_external_data: bool = False,
    update_metadata: bool = True,
    respect_dryrun_skip: bool = True,
    sql_dir=ConfigLoader.get("default", "sql_dir"),
    credentials=None,
    id_token=None,
) -> None:
    """Deploy a query to a destination."""
    if respect_dryrun_skip and str(artifact_file) in DryRun.skipped_files():
        raise SkippedDeployException(f"Dry run skipped for {artifact_file}.")

    metadata = None
    try:
        if artifact_file.name == METADATA_FILE:
            metadata = Metadata.from_file(artifact_file)
        else:
            metadata = Metadata.of_query_file(artifact_file)

        if metadata.external_data:
            if artifact_file.suffix == ".sql" or artifact_file.name == "query.py":
                raise FailedDeployException(
                    f"Invalid metadata: {artifact_file} has both a SQL file and "
                    f"external data config"
                )
            if skip_external_data:
                raise SkippedExternalDataException(
                    f"Skipping deploy of external data table {artifact_file}"
                )

        if (
            metadata.scheduling
            and "destination_table" in metadata.scheduling
            and metadata.scheduling["destination_table"] is None
        ):
            raise SkippedDeployException(
                f"Skipping deploy for {artifact_file}, null destination_table configured."
            )
    except FileNotFoundError:
        log.warning(f"No metadata found for {artifact_file}.")

    table_name = artifact_file.parent.name
    dataset_name = artifact_file.parent.parent.name
    project_name = artifact_file.parent.parent.parent.name

    if destination_table is None:
        destination_table = f"{project_name}.{dataset_name}.{table_name}"

    existing_schema_path = artifact_file.parent / SCHEMA_FILE
    try:
        existing_schema = Schema.from_schema_file(existing_schema_path)
    except Exception as e:  # TODO: Raise/catch more specific exception
        raise SkippedDeployException(f"Schema missing for {artifact_file}.") from e

    client = bigquery.Client(credentials=credentials)
    if not force and str(artifact_file).endswith("query.sql"):
        query_schema = Schema.from_query_file(
            artifact_file,
            use_cloud_function=use_cloud_function,
            respect_skip=respect_dryrun_skip,
            sql_dir=sql_dir,
            client=client,
            id_token=id_token if not use_cloud_function or id_token else get_id_token(),
        )
        if not existing_schema.equal(query_schema):
            raise FailedDeployException(
                f"Query {artifact_file} does not match "
                f"schema in {existing_schema_path}. "
                f"To update the local schema file, "
                f"run `./bqetl query schema update "
                f"{dataset_name}.{table_name}`",
            )

    try:
        table = client.get_table(destination_table)
    except NotFound:
        table = bigquery.Table(destination_table)
    table.schema = existing_schema.to_bigquery_schema()

    if metadata and metadata.external_data:
        attach_external_data_config(artifact_file, table)

    if update_metadata:
        attach_metadata(artifact_file, table)

    _create_or_update(client, table, skip_existing)


def _create_or_update(
    client: bigquery.Client,
    table: bigquery.Table,
    skip_existing: bool = False,
) -> None:
    if table.created:
        if skip_existing:
            raise SkippedDeployException(f"{table} already exists.")
        log.info(f"{table} already exists, updating.")
        try:
            client.update_table(
                table,
                [
                    "schema",
                    "friendly_name",
                    "description",
                    "time_partitioning",
                    "clustering_fields",
                    "labels",
                ],
            )
        except Exception as e:
            raise FailedDeployException(f"Unable to update table {table}: {e}") from e
        log.info(f"{table} updated.")
    else:
        try:
            client.create_table(table)
        except Exception as e:
            raise FailedDeployException(f"Unable to create table {table}: {e}") from e
        log.info(f"{table} created.")



import os
import types
from datetime import date, timedelta
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from click.testing import CliRunner

from bigquery_etl.cli.query import (
    NBR_DAYS_RETAINED,
    backfill,
    create,
    deploy,
    info,
    materialized_view_has_changes,
    paths_matching_name_pattern,
    schedule,
)
from bigquery_etl.metadata.publish_metadata import attach_metadata


class TestQuery:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def create_test_dag(self, dag_name):
        os.mkdir("dags")

        dag_conf = {
            f"{dag_name}": {
                "schedule_interval": "daily",
                "default_args": {
                    "owner": "test@example.com",
                    "start_date": "2020-03-29",
                    "email": ["test@example.org"],
                    "retries": 1,
                },
            }
        }

        with open("dags.yaml", "w") as f:
            f.write(yaml.dump(dag_conf))

    def test_create_invalid_path(self, runner):
        with runner.isolated_filesystem():
            with open("foo.txt", "w") as f:
                f.write("")
            result = runner.invoke(create, ["test.query_v1", "--sql_dir=foo.txt"])
            assert result.exit_code == 2

    def test_create_invalid_query_name(self, runner):
        with runner.isolated_filesystem():
            result = runner.invoke(create, ["invalid_query_name"])
            assert result.exit_code == 2

    def test_create_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query", "--no_schedule"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod") == ["test"]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
            ]
            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )

    def test_create_query_in_default_dag(self, runner):
        with runner.isolated_filesystem():
            self.create_test_dag("bqetl_default")
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod") == ["test"]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
            ]
            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml"
            ) as file:
                exists = "dag_name: bqetl_default" in file.read()
                assert exists

    def test_create_query_in_named_dag(self, runner):
        with runner.isolated_filesystem():
            self.create_test_dag("bqetl_test")
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query", "--dag=bqetl_test"])
            assert result.exit_code == 0
            assert os.listdir("sql/moz-fx-data-shared-prod") == ["test"]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
            ]
            assert "query.sql" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            assert "metadata.yaml" in os.listdir(
                "sql/moz-fx-data-shared-prod/test/test_query_v1"
            )
            with open(
                "sql/moz-fx-data-shared-prod/test/test_query_v1/metadata.yaml"
            ) as file:
                exists = "dag_name: bqetl_test" in file.read()
                assert exists

    def test_create_query_with_version(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            result = runner.invoke(create, ["test.test_query_v4", "--no_schedule"])
            assert result.exit_code == 0
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query_v4",
            ]

    def test_create_derived_query_with_view(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test_derived")
            result = runner.invoke(create, ["test_derived.test_query", "--no_schedule"])
            assert result.exit_code == 0
            assert "test_derived" in os.listdir("sql/moz-fx-data-shared-prod")
            assert "test" in os.listdir("sql/moz-fx-data-shared-prod")
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test_derived")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
            ]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query",
            ]
            assert os.listdir("sql/moz-fx-data-shared-prod/test/test_query") == [
                "view.sql"
            ]

    def test_create_derived_query_with_existing_view(self, runner):
        with runner.isolated_filesystem():
            project_dir = Path("sql/moz-fx-data-shared-prod")
            existing_query = (
                project_dir / "test_derived" / "test_query_v1" / "query.sql"
            )
            existing_query.parent.mkdir(parents=True)
            existing_query.touch()

            existing_view = project_dir / "test" / "test_query" / "view.sql"
            existing_view.parent.mkdir(parents=True)
            existing_view.touch()

            result = runner.invoke(
                create, ["test_derived.test_query_v2", "--no_schedule"]
            )
            assert result.exit_code == 0
            assert "test_derived" in os.listdir("sql/moz-fx-data-shared-prod")
            assert "test" in os.listdir("sql/moz-fx-data-shared-prod")
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test_derived")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
                "test_query_v2",
            ]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query",
            ]
            assert os.listdir("sql/moz-fx-data-shared-prod/test/test_query") == [
                "view.sql"
            ]
            assert existing_view.read_text() == ""

    def test_create_query_as_derived(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod")
            os.mkdir("sql/moz-fx-data-shared-prod/test_derived")
            os.mkdir("sql/moz-fx-data-shared-prod/test")
            result = runner.invoke(create, ["test.test_query", "--no_schedule"])
            assert result.exit_code == 0
            assert "test_derived" in os.listdir("sql/moz-fx-data-shared-prod")
            assert "test" in os.listdir("sql/moz-fx-data-shared-prod")
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test_derived")) == [
                "dataset_metadata.yaml",
                "test_query_v1",
            ]
            assert sorted(os.listdir("sql/moz-fx-data-shared-prod/test")) == [
                "dataset_metadata.yaml",
                "test_query",
            ]
            assert os.listdir("sql/moz-fx-data-shared-prod/test/test_query") == [
                "view.sql"
            ]

    def test_schedule_invalid_path(self, runner):
        with runner.isolated_filesystem():
            result = runner.invoke(schedule, ["/test/query_v1"])
            assert result.exit_code == 2

    def test_schedule_query_non_existing_dag(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/test/query_v1")
            open("sql/moz-fx-data-shared-prod/test/query_v1/query.sql", "a").close()
            result = runner.invoke(schedule, ["test.query_v1", "--dag=foo"])
            assert result.exit_code == 1

    def test_schedule_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            os.mkdir("dags")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            dag_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.com",
                        "start_date": "2020-03-29",
                        "email": ["test@example.org"],
                        "retries": 1,
                    },
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dag_conf))

            result = runner.invoke(
                schedule, ["telemetry_derived.query_v1", "--dag=bqetl_test"]
            )

            print(result.output)
            assert result.exit_code == 0

    def test_reschedule_query(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            os.mkdir("dags")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            dag_conf = {
                "bqetl_test": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.com",
                        "start_date": "2020-03-29",
                        "email": ["test@example.org"],
                        "retries": 1,
                    },
                }
            }

            with open("dags.yaml", "w") as f:
                f.write(yaml.dump(dag_conf))

            result = runner.invoke(schedule, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0

    def test_query_info(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            result = runner.invoke(info, ["*.telemetry_derived.query_v1"])
            assert result.exit_code == 0
            assert "No metadata" in result.output
            assert "path:" in result.output

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            result = runner.invoke(info, ["*.telemetry_derived.query_v1"])
            assert result.exit_code == 0
            assert "No metadata" not in result.output
            assert "description" in result.output
            assert "dag_name: bqetl_test" in result.output

    def test_info_name_pattern(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.mkdir("sql/moz-fx-data-shared-prod/telemetry_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-shared-prod/foo_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/foo_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            result = runner.invoke(info, ["*.query_*"])
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" in result.output
            assert "telemetry_derived.query_v1" in result.output

            result = runner.invoke(
                info, ["foo_derived.*", "--project-id=moz-fx-data-shared-prod"]
            )
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" not in result.output
            assert "telemetry_derived.query_v1" not in result.output

            result = runner.invoke(info, ["*.query_v2"])
            assert result.exit_code == 0
            assert "foo_derived.query_v2" in result.output
            assert "telemetry_derived.query_v2" in result.output
            assert "telemetry_derived.query_v1" not in result.output

    def testpaths_matching_name_pattern(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.mkdir("sql/moz-fx-data-shared-prod/telemetry_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-shared-prod/foo_derived/query_v2")
            with open(
                "sql/moz-fx-data-shared-prod/foo_derived/query_v2/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            os.makedirs("sql/moz-fx-data-test-project/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-test-project/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            assert len(paths_matching_name_pattern("*", "sql/", None)) == 4
            assert (
                len(
                    paths_matching_name_pattern(
                        "*.sql", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    paths_matching_name_pattern(
                        "test", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    paths_matching_name_pattern(
                        "foo_derived", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 0
            )
            assert (
                len(
                    paths_matching_name_pattern(
                        "foo_derived*", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 1
            )
            assert len(paths_matching_name_pattern("*query*", "sql/", None)) == 4
            assert (
                len(
                    paths_matching_name_pattern(
                        "foo_derived.query_v2", "sql/", "moz-fx-data-shared-prod"
                    )
                )
                == 1
            )

            assert (
                len(
                    paths_matching_name_pattern(
                        "telemetry_derived.query_v1", "sql/", "moz-fx-data-test-project"
                    )
                )
                == 1
            )

            assert (
                len(
                    paths_matching_name_pattern(
                        "moz-fx-data-test-project.telemetry_derived.query_v1",
                        "sql/",
                        None,
                    )
                )
                == 1
            )

            assert (
                len(
                    paths_matching_name_pattern(
                        "moz-fx-data-test-project.telemetry_derived.*", "sql/", None
                    )
                )
                == 1
            )

    def test_attach_metadata_labels(self, runner):
        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
                "labels": {"test": 123, "foo": "abc", "review_bugs": [1234, 1254]},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            table = types.SimpleNamespace()
            attach_metadata(
                Path(
                    "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql"
                ),
                table,
            )

            assert "test" in table.labels
            assert table.labels["test"] == "123"
            assert "foo" in table.labels
            assert table.labels["foo"] == "abc"
            assert "review_bugs" not in table.labels

    def test_query_backfill_with_offset(self, runner):
        with (
            runner.isolated_filesystem(),
            # Mock client to avoid NotFound
            patch("google.cloud.bigquery.Client", autospec=True),
            patch("subprocess.check_call", autospec=True) as check_call,
        ):
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write(
                    "SELECT DATE('2021-01-01') as submission_date WHERE submission_date = @submission_date"
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test", "date_partition_offset": -2},
                "bigquery": {"time_partitioning": {"type": "day"}},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            result = runner.invoke(
                backfill,
                [
                    "telemetry_derived.query_v1",
                    "--project_id=moz-fx-data-shared-prod",
                    "--start_date=2021-01-05",
                    "--end_date=2021-01-09",
                    "--exclude=2021-01-06",
                    "--parallelism=0",
                    "--billing-project=backfill-project",
                    "--override-retention-range-limit",
                ],
            )

            assert result.exit_code == 0

            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}"
                for day in (3, 5, 6, 7)
            ]

            assert check_call.call_count == 4
            for call in check_call.call_args_list:
                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params

    def test_query_backfill_with_scheduling_overrides(self, runner):
        with (
            runner.isolated_filesystem(),
            # Mock client to avoid NotFound
            patch("google.cloud.bigquery.Client", autospec=True),
            patch("subprocess.check_call", autospec=True) as check_call,
        ):
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write(
                    "SELECT DATE('2021-01-01') as submission_date WHERE submission_date = @submission_date"
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {
                    "dag_name": "bqetl_test",
                    "date_partition_parameter": None,
                    "parameters": [
                        "submission_date:DATE:{{(execution_date - macros.timedelta(hours=1)).strftime('%Y-%m-%d')}}",
                    ],
                },
                "bigquery": {"time_partitioning": {"type": "day"}},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(
                backfill,
                [
                    "telemetry_derived.query_v1",
                    "--project_id=moz-fx-data-shared-prod",
                    "--start_date=2021-01-05",
                    "--end_date=2021-01-06",
                    "--parallelism=0",
                    """--scheduling_overrides={"parameters": ["test:INT64:30"], "date_partition_parameter": "submission_date"}""",
                    "--override-retention-range-limit",
                ],
            )
            assert result.exit_code == 0

            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}" for day in (5, 6)
            ]
            assert check_call.call_count == 2

            for call in check_call.call_args_list:
                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params

                test_params = [arg for arg in call.args[0] if "--parameter=test" in arg]
                assert len(test_params) == 1
                assert test_params[0] == "--parameter=test:INT64:30"

    def test_query_backfill_unpartitioned_with_parameters(self, runner):
        with (
            runner.isolated_filesystem(),
            # Mock client to avoid NotFound
            patch("google.cloud.bigquery.Client", autospec=True),
            patch("subprocess.check_call", autospec=True) as check_call,
        ):
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write(
                    "SELECT DATE('2021-01-01') as submission_date WHERE submission_date = @submission_date"
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {
                    "dag_name": "bqetl_test",
                    "date_partition_parameter": None,
                    "parameters": [
                        "submission_date:DATE:{{ds}}",
                        "conversion_window:INT64:30",
                    ],
                },
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))

            result = runner.invoke(
                backfill,
                [
                    "telemetry_derived.query_v1",
                    "--project_id=moz-fx-data-shared-prod",
                    "--start_date=2021-01-05",
                    "--end_date=2021-01-09",
                    "--parallelism=0",
                    "--override-retention-range-limit",
                ],
            )

            expected_destination_arg = (
                "--destination_table=moz-fx-data-shared-prod:telemetry_derived.query_v1"
            )
            expected_submission_date_params = [
                f"--parameter=submission_date:DATE:2021-01-0{day}"
                for day in range(5, 10)
            ]

            assert result.exit_code == 0

            assert check_call.call_count == 5
            for call in check_call.call_args_list:
                destination_table_params = [
                    arg for arg in call.args[0] if "--destination_table" in arg
                ]
                assert len(destination_table_params) == 1
                assert destination_table_params[0] == expected_destination_arg

                submission_date_params = [
                    arg for arg in call.args[0] if "--parameter=submission_date" in arg
                ]
                assert len(submission_date_params) == 1
                assert submission_date_params[0] in expected_submission_date_params

                conversion_params = [
                    arg
                    for arg in call.args[0]
                    if "--parameter=conversion_window" in arg
                ]
                assert len(conversion_params) == 1
                assert conversion_params[0] == "--parameter=conversion_window:INT64:30"

    @patch("bigquery_etl.cli.query.get_credentials")
    @patch("bigquery_etl.cli.query.get_id_token")
    @patch("bigquery_etl.cli.query.deploy_table")
    def test_deploy(
        self, mock_deploy_table, mock_get_id_token, mock_get_credentials, runner
    ):
        mock_deploy_table.return_value = None
        mock_get_id_token.return_value = None
        mock_get_credentials.return_value = None

        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
                "labels": {"test": 123, "foo": "abc", "review_bugs": [1234, 1254]},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(deploy, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0
            mock_deploy_table.assert_called_with(
                Path(
                    "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql"
                ),
                destination_table=None,
                force=False,
                use_cloud_function=True,
                skip_existing=False,
                skip_external_data=False,
                respect_dryrun_skip=True,
                sql_dir="sql/",
                credentials=None,
                id_token=None,
            )
            mock_get_id_token.assert_called_once()
            mock_get_credentials.assert_called_once()

    @patch("bigquery_etl.cli.query.get_credentials")
    @patch("bigquery_etl.cli.query.get_id_token")
    @patch("bigquery_etl.cli.query.deploy_table")
    def test_deploy_schema(
        self, mock_deploy_table, mock_get_id_token, mock_get_credentials, runner
    ):
        mock_deploy_table.return_value = None
        mock_get_id_token.return_value = None
        mock_get_credentials.return_value = None

        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/schema.yaml",
                "w",
            ) as f:
                f.write(
                    """
                fields:
                - name: x
                  type: INTEGER
                  mode: NULLABLE
                """
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
                "labels": {"test": 123, "foo": "abc", "review_bugs": [1234, 1254]},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(deploy, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0
            mock_deploy_table.assert_called_with(
                Path(
                    "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml"
                ),
                destination_table=None,
                force=False,
                use_cloud_function=True,
                skip_existing=False,
                skip_external_data=False,
                respect_dryrun_skip=True,
                sql_dir="sql/",
                credentials=None,
                id_token=None,
            )
            mock_get_id_token.assert_called_once()
            mock_get_credentials.assert_called_once()

    @patch("bigquery_etl.cli.query.get_credentials")
    @patch("bigquery_etl.cli.query.get_id_token")
    @patch("bigquery_etl.cli.query.deploy_table")
    def test_deploy_schema_no_duplicate(
        self, mock_deploy_table, mock_get_id_token, mock_get_credentials, runner
    ):
        mock_deploy_table.return_value = None
        mock_get_id_token.return_value = None
        mock_get_credentials.return_value = None

        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/schema.yaml",
                "w",
            ) as f:
                f.write(
                    """
                fields:
                - name: x
                  type: INTEGER
                  mode: NULLABLE
                """
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
                "labels": {"test": 123, "foo": "abc", "review_bugs": [1234, 1254]},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(deploy, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0
            mock_deploy_table.assert_called_once()
            mock_deploy_table.assert_called_with(
                Path(
                    "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql"
                ),
                destination_table=None,
                force=False,
                use_cloud_function=True,
                skip_existing=False,
                skip_external_data=False,
                respect_dryrun_skip=True,
                sql_dir="sql/",
                credentials=None,
                id_token=None,
            )
            mock_get_id_token.assert_called_once()
            mock_get_credentials.assert_called_once()

    @patch("bigquery_etl.cli.query.get_credentials")
    @patch("bigquery_etl.cli.query.get_id_token")
    @patch("bigquery_etl.cli.query.deploy_table")
    def test_prevent_deploy_for_views(
        self, mock_deploy_table, mock_get_id_token, mock_get_credentials, runner
    ):
        mock_deploy_table.return_value = None
        mock_get_id_token.return_value = None
        mock_get_credentials.return_value = None

        with runner.isolated_filesystem():
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")
            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/view.sql", "w"
            ) as f:
                f.write("SELECT 1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/schema.yaml",
                "w",
            ) as f:
                f.write(
                    """
                fields:
                - name: x
                  type: INTEGER
                  mode: NULLABLE
                """
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {"dag_name": "bqetl_test"},
                "labels": {"test": 123, "foo": "abc", "review_bugs": [1234, 1254]},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(deploy, ["telemetry_derived.query_v1"])

            assert result.exit_code == 0
            mock_deploy_table.assert_not_called()

    def test_materialized_view_has_changes(self):
        deployed_sql = " SELECT  * FROM project.dataset.table "

        assert not materialized_view_has_changes(
            deployed_sql,
            """
            CREATE MATERIALIZED VIEW project.dataset.view AS SELECT * FROM
            project.dataset.table
            """,
        )
        assert not materialized_view_has_changes(
            deployed_sql,
            """
            CREATE MATERIALIZED VIEW IF NOT EXISTS project.dataset.view AS SELECT * FROM
            project.dataset.table
            """,
        )
        assert materialized_view_has_changes(
            deployed_sql,
            """
            CREATE MATERIALIZED VIEW IF NOT EXISTS project.dataset.view AS SELECT * FROM
            project.dataset.table WHERE TRUE
            """,
        )
        assert not materialized_view_has_changes(
            deployed_sql,
            """
            CREATE OR REPLACE MATERIALIZED VIEW
            project.dataset.view AS SELECT * FROM
            project.dataset.table
            """,
        )

    # Check that if you try to run a backfill for a start date < retention days and retention = False, it exits
    def test_query_backfill_with_scheduling_overrides_outside_retention_limit(
        self, runner
    ):
        with (
            runner.isolated_filesystem(),
            # Mock client to avoid NotFound
            patch("google.cloud.bigquery.Client", autospec=True),
        ):
            os.makedirs("sql/moz-fx-data-shared-prod/telemetry_derived/query_v1")

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/query.sql", "w"
            ) as f:
                f.write(
                    "SELECT DATE('2021-01-01') as submission_date WHERE submission_date = @submission_date"
                )

            metadata_conf = {
                "friendly_name": "test",
                "description": "test",
                "owners": ["test@example.org"],
                "scheduling": {
                    "dag_name": "bqetl_test",
                    "date_partition_parameter": None,
                    "parameters": [
                        "submission_date:DATE:{{(execution_date - macros.timedelta(hours=1)).strftime('%Y-%m-%d')}}",
                    ],
                },
                "bigquery": {"time_partitioning": {"type": "day"}},
            }

            with open(
                "sql/moz-fx-data-shared-prod/telemetry_derived/query_v1/metadata.yaml",
                "w",
            ) as f:
                f.write(yaml.dump(metadata_conf))
            result = runner.invoke(
                backfill,
                [
                    "telemetry_derived.query_v1",
                    "--project_id=moz-fx-data-shared-prod",
                    f"--start_date={str(date.today() - timedelta(days=10+NBR_DAYS_RETAINED))}",
                    f"--end_date={str(date.today() - timedelta(days=10))}",
                    "--parallelism=0",
                    """--scheduling_overrides={"parameters": ["test:INT64:30"], "date_partition_parameter": "submission_date"}""",
                ],
            )
            assert result.exit_code == 1



from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

from google.cloud import bigquery

from bigquery_etl.routine import parse_routine, publish_routines

TEST_DIR = Path(__file__).parent.parent


class TestPublishRoutine:
    udf_dir = TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project" / "udf"

    @mock.patch("google.cloud.bigquery.Client")
    def test_publish_routine_with_description(self, mock_client):
        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_shift_28_bits_one_day" / "udf.sql"
        )
        mock_client.query = MagicMock()
        publish_routines.publish_routine(
            raw_routine, mock_client, "test-project", "", "", [], False
        )
        query = (
            "CREATE OR REPLACE FUNCTION udf.test_shift_28_bits_one_day(x INT64)"
            + " AS (\n  IFNULL((x << 1) & udf.test_bitmask_lowest_28(), 0)\n)"
            + 'OPTIONS(description="Shift input bits one day left and drop any bits'
            + ' beyond 28 days.");'
        )
        job_config = bigquery.QueryJobConfig(dry_run=False)
        assert [
            (
                c.args,
                {
                    k: (v.to_api_repr() if k == "job_config" else v)
                    for k, v in c.kwargs.items()
                },
            )
            for c in mock_client.query.call_args_list
        ] == [((query,), {"job_config": job_config.to_api_repr()})]

    @mock.patch("google.cloud.bigquery.Client")
    def test_publish_js_udf_with_description(self, mock_client):
        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_js_udf" / "udf.sql"
        )
        mock_client.query = MagicMock()
        publish_routines.publish_routine(
            raw_routine, mock_client, "test-project", "", "", [], False
        )
        query = (
            "CREATE OR REPLACE FUNCTION udf.test_js_udf(input BYTES)\nRETURNS "
            + "STRING DETERMINISTIC\nLANGUAGE js\nAS\n "
            + ' """return 1;"""\nOPTIONS(description="Some description",'
            + 'library = "gs:///script.js");'
        )
        job_config = bigquery.QueryJobConfig(dry_run=False)
        assert [
            (
                c.args,
                {
                    k: (v.to_api_repr() if k == "job_config" else v)
                    for k, v in c.kwargs.items()
                },
            )
            for c in mock_client.query.call_args_list
        ] == [((query,), {"job_config": job_config.to_api_repr()})]



"""Utilities for the GLAM module."""

import json
import subprocess
from collections import namedtuple
from itertools import combinations
from typing import List, Tuple

from mozilla_schema_generator.glean_ping import GleanPing

CustomDistributionMeta = namedtuple(
    "CustomDistributionMeta",
    ["name", "type", "range_min", "range_max", "bucket_count", "histogram_type"],
)


def run(command, **kwargs) -> str:
    """Return the result of stdout and raise on subprocess error."""
    if isinstance(command, list):
        args = command
    elif isinstance(command, str):
        args = command.split()
    else:
        raise RuntimeError(f"run command is invalid: {command}")

    try:
        res = (
            subprocess.run(args, stdout=subprocess.PIPE, check=True, **kwargs)
            .stdout.decode()
            .strip()
        )
    except subprocess.CalledProcessError as e:
        print(e.output.decode())
        raise e

    return res


def get_schema(table: str, project: str = "moz-fx-data-shared-prod"):
    """Return the dictionary representation of the BigQuery table schema.

    This returns types in the legacy SQL format.
    """
    process = subprocess.Popen(
        ["bq", "show", "--schema", "--format=json", f"{project}:{table}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    if process.returncode > 0:
        raise Exception(
            f"Call to bq exited non-zero: {process.returncode}", stdout, stderr
        )
    return json.loads(stdout)


def ping_type_from_table(qualified_table):
    """Return the name of a ping as defined in mozilla-pipeline-schemas.

    Example: org_mozilla_fenix_stable.deletion_request_v1 -> deletion-request
    """
    table_id = qualified_table.split(".")[-1]
    ping_name = table_id.rsplit("_", 1)[0]
    return ping_name.replace("_", "-")


def get_custom_distribution_metadata(product_name) -> List[CustomDistributionMeta]:
    """Get metadata for reconstructing custom distribution buckets in Glean metrics."""
    # GleanPing.get_repos -> List[Tuple[name: str, app_id: str]]
    glean = GleanPing(dict(name=product_name, app_id=product_name))
    probes = glean.get_probes()

    custom = []
    for probe in probes:
        # We use endswith here to accommodate for types prefixed with "labeled"
        if not probe.get_type().endswith("custom_distribution"):
            continue
        meta = CustomDistributionMeta(
            probe.get_name(),
            probe.get_type(),
            probe.get("range_min"),
            probe.get("range_max"),
            probe.get("bucket_count"),
            probe.get("histogram_type"),
        )
        custom.append(meta)

    return custom


def compute_datacube_groupings(
    attributes: List[str], fixed_attributes: List[str] = []
) -> List[List[Tuple[str, bool]]]:
    """Generate the combinations of attributes to be computed.

    These are the combinations that are available to the frontend. Some
    dimensions may be fixed and always required.
    """
    max_combinations = len(attributes)
    result = []
    for subset_size in reversed(range(max_combinations + 1)):
        for grouping in combinations(attributes, subset_size):
            if len(set(fixed_attributes) - set(grouping)) > 0:
                # the fixed attributes are always a subset in the grouping
                continue
            select_expr = []
            for attribute in attributes:
                select_expr.append((attribute, attribute in grouping))
            result.append(select_expr)
    return result


if __name__ == "__main__":
    from pprint import PrettyPrinter

    pp = PrettyPrinter()
    pp.pprint(get_custom_distribution_metadata("fenix"))



"""Generate templated views."""

from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from functools import partial
from pathlib import Path

from jinja2 import Environment, PackageLoader, TemplateNotFound
from jsonschema import validate

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.glam import models


class QueryType:
    """Types of queries in the template folder."""

    VIEW = "view"
    INIT = "init"
    TABLE = "query"


@dataclass
class TemplateResult:
    """Results of templating a query."""

    table_id: str
    query_type: QueryType
    query_text: str


def from_template(
    query_type: QueryType,
    template_name: str,
    environment: Environment,
    args: Namespace,
    dataset_path: Path,
    query_name_prefix=None,
    **kwargs,
) -> TemplateResult:
    """Fill in templates and write them to disk."""
    if query_type == QueryType.INIT:
        template = environment.get_template(f"{template_name}.init.sql")
    else:
        template = environment.get_template(f"{template_name}.sql")

    template_filename = template_name.split("__")[-1]
    if query_name_prefix:
        table_id = f"{args.prefix}__{query_name_prefix}_{template_filename}"
    else:
        table_id = f"{args.prefix}__{template_filename}"

    # replaces the header, if it exists
    kwargs["header"] = f"-- {query_type} for {table_id};"

    # create the directory for the view
    (dataset_path / table_id).mkdir(exist_ok=True)
    view_path = dataset_path / table_id / f"{query_type}.sql"

    # write the query with appropriate variables
    query_text = reformat(template.render(**{**vars(args), **kwargs}))

    print(f"generated {view_path}")
    with view_path.open("w") as fp:
        print(query_text, file=fp)

    return TemplateResult(table_id, query_type, query_text)


def main():
    """Generate GLAM ETL queries."""
    parser = ArgumentParser(description=main.__doc__)
    parser.add_argument("--prefix")
    parser.add_argument("--project", default="glam-fenix-dev")
    parser.add_argument("--dataset", default="glam_etl")
    parser.add_argument("--sql-root", default="sql/")
    parser.add_argument("--daily-view-only", action="store_true", default=False)
    parser.add_argument("--use-sample-id", action="store_true", default=False)
    args = parser.parse_args()

    env = Environment(loader=PackageLoader("bigquery_etl", "glam/templates"))

    dataset_path = Path(args.sql_root) / args.project / args.dataset
    if not dataset_path.is_dir():
        raise NotADirectoryError(f"path to {dataset_path} not found")

    # curry functions for convenience
    template = partial(
        from_template, environment=env, dataset_path=dataset_path, args=args
    )
    view = partial(template, QueryType.VIEW)
    table = partial(template, QueryType.TABLE)
    init = partial(template, QueryType.INIT)

    # If this is a logical app id, generate it. Assert that the daily view for
    # the app exists. This assumes that both scalar and histogram aggregates
    # exist and will break down in the case where a glean app only contains one
    # of the scalar or histogram view.
    for daily_view in [
        "view_clients_daily_scalar_aggregates_v1",
        "view_clients_daily_histogram_aggregates_v1",
    ]:
        try:
            view(f"logical_app_id/{args.prefix}__{daily_view}")
        except TemplateNotFound:
            print(f"{args.prefix} is not a logical app id")
            # generate the view for the app id directly
            view(daily_view)

        if not (dataset_path / f"{args.prefix}__{daily_view}").is_dir():
            raise ValueError(f"missing {daily_view}")

    # exit early if we're only generating a daily view
    if args.daily_view_only:
        return

    # Supported fenix/firefox for android products. These are logical ids that
    # are formed from the union of several app_ids (sometimes across date
    # boundaries).
    config_schema = {
        "type": "object",
        "additionalProperties": {
            "type": "object",
            "properties": {
                "build_date_udf": {
                    "type": "string",
                    "description": (
                        "The fully qualified path to a UDF that accepts the "
                        "client_info.app_build_id field and returns a datetime."
                    ),
                },
                "filter_version": {
                    "type": "boolean",
                    "description": (
                        "Whether the integer extracted from the "
                        "client_info.app_display_version should be used to "
                        "filter incremental aggregates."
                    ),
                },
                "num_versions_to_keep": {
                    "type": "integer",
                    "minimum": 1,
                    "description": "The number of versions to keep.",
                },
                "total_users": {
                    "type": "integer",
                    "minimum": 1,
                    "description": "The number of users to filter the data on.",
                },
                "minimum_client_count": {
                    "type": "integer",
                    "minimum": 0,
                    "description": "The minimum client count for each build id."
                    "We generally want this to be roughly 0.5% of WAU."
                    "For context see https://github.com/mozilla/glam/issues/1575#issuecomment-946880387",  # noqa E501
                },
            },
            "required": [
                "build_date_udf",
                "filter_version",
                "num_versions_to_keep",
                "total_users",
            ],
        },
    }
    config = {
        "org_mozilla_fenix_glam_nightly": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 10,
            "minimum_client_count": 800,
        },
        "org_mozilla_fenix_glam_beta": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 10,
            "minimum_client_count": 2000,
        },
        "org_mozilla_fenix_glam_release": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 10,
            "minimum_client_count": 90000,
        },
        "firefox_desktop_glam_nightly": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 10,
            "minimum_client_count": 100,
        },
        "firefox_desktop_glam_beta": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 100,
            "minimum_client_count": 1000,
        },
        "firefox_desktop_glam_release": {
            "build_date_udf": "mozfun.glam.build_hour_to_datetime",
            "filter_version": True,
            "num_versions_to_keep": 3,
            "total_users": 100,
            "minimum_client_count": 100000,
        },
    }

    channel_prefixes = {
        "firefox_desktop_glam_nightly": "nightly",
        "firefox_desktop_glam_beta": "beta",
        "firefox_desktop_glam_release": "release",
        "org_mozilla_fenix_glam_nightly": "nightly",
        "org_mozilla_fenix_glam_beta": "beta",
        "org_mozilla_fenix_glam_release": "release",
    }
    validate(instance=config, schema=config_schema)

    if not config.get(args.prefix, {}).get("build_date_udf"):
        raise ValueError(f"build date udf for {args.prefix} was not found")

    [
        table(
            "latest_versions_v1",
            **dict(app_id_channel=(f"'{channel_prefixes[args.prefix]}'")),
        ),
        init(
            "clients_scalar_aggregates_v1",
            **models.clients_scalar_aggregates(
                source_table=(
                    f"glam_etl.{args.prefix}__view_clients_daily_scalar_aggregates_v1"
                ),
                destination_table=(
                    f"glam_etl.{args.prefix}__clients_scalar_aggregates_v1"
                ),
            ),
        ),
        table(
            "clients_scalar_aggregates_v1",
            **models.clients_scalar_aggregates(
                source_table=(
                    f"glam_etl.{args.prefix}__view_clients_daily_scalar_aggregates_v1"
                ),
                destination_table=(
                    f"glam_etl.{args.prefix}__clients_scalar_aggregates_v1"
                ),
                **config[args.prefix],
            ),
        ),
        init(
            "clients_histogram_aggregates_v1",
            **models.clients_histogram_aggregates(parameterize=True),
        ),
        table(
            "clients_histogram_aggregates_v1",
            **models.clients_histogram_aggregates(
                parameterize=True, **config[args.prefix]
            ),
        ),
        table(
            "scalar_bucket_counts_v1",
            **models.scalar_bucket_counts(
                source_table=f"glam_etl.{args.prefix}__clients_scalar_aggregates_v1",
                **config[args.prefix],
            ),
        ),
        table(
            "histogram_bucket_counts_v1",
            **models.histogram_bucket_counts(
                source_table=f"glam_etl.{args.prefix}__clients_histogram_aggregates_v1",
                **config[args.prefix],
            ),
            channel=channel_prefixes[args.prefix],
            use_sample_id=args.use_sample_id,
        ),
        table(
            "probe_counts_v1",
            query_name_prefix="scalar",
            **models.probe_counts(
                source_table=f"glam_etl.{args.prefix}__scalar_bucket_counts_v1",
                is_scalar=True,
            ),
            channel=channel_prefixes[args.prefix],
        ),
        table(
            "probe_counts_v1",
            query_name_prefix="histogram",
            **models.probe_counts(
                source_table=f"glam_etl.{args.prefix}__histogram_bucket_counts_v1",
                is_scalar=False,
            ),
            channel=channel_prefixes[args.prefix],
        ),
        view("view_probe_counts_v1"),
        view("view_user_counts_v1", **models.user_counts()),
        view(
            "view_sample_counts_v1",
            **models.sample_counts(),
            channel=channel_prefixes[args.prefix],
        ),
        table("extract_user_counts_v1", **config[args.prefix]),
        table("extract_probe_counts_v1", **config[args.prefix]),
    ]


if __name__ == "__main__":
    main()



"""Queue for balancing jobs across billing projects."""

import asyncio
from contextlib import contextmanager
from queue import Queue

from google.cloud import bigquery
from requests import adapters


class ClientQueue:
    """Queue for balancing jobs across billing projects.

    Also provides default_client for use in operations that do not need to be
    distributed across projects and that may be in excess of parallelism, such
    as copying results from a subset of queries before all queries have
    finished.
    """

    def __init__(self, billing_projects, parallelism, connection_pool_max_size=None):
        """Initialize.

        connection_pool_max_size sets the pool size in the HTTP adapter of each client in
        the queue. This allows more concurrent requests when a client is shared across threads.
        See https://cloud.google.com/bigquery/docs/python-libraries#troubleshooting_connection_pool_errors
        Increasing connection_pool_max_size will also increase memory usage.
        """
        clients = [bigquery.Client(project) for project in billing_projects]

        if connection_pool_max_size is not None:
            for client in clients:
                adapter = adapters.HTTPAdapter(
                    pool_connections=connection_pool_max_size,
                    pool_maxsize=connection_pool_max_size,
                )
                client._http.mount("https://", adapter)
                client._http._auth_request.session.mount("https://", adapter)

        self.default_client = clients[0]
        self._q = Queue(parallelism)
        for i in range(parallelism):
            self._q.put(clients[i % len(clients)])

    @contextmanager
    def client(self):
        """Context manager for using a client from the queue."""
        client = self._q.get_nowait()
        try:
            yield client
        finally:
            self._q.put_nowait(client)

    def with_client(self, func, *args):
        """Run func with a client from the queue."""
        with self.client() as client:
            return func(client, *args)

    async def async_with_client(self, executor, func, *args):
        """Run func asynchronously in executor."""
        return await asyncio.get_running_loop().run_in_executor(
            executor, self.with_client, func, *args
        )



"""
Machinery for parsing UDFs and tests defined in .sql files.

This should eventually be refactored to a more general library for
parsing UDF dependencies in queries as well.
"""

import os
import re
from pathlib import Path
from typing import List

import attr
import sqlparse
import yaml

from bigquery_etl.config import ConfigLoader
from bigquery_etl.metadata.parse_metadata import METADATA_FILE
from bigquery_etl.util.common import render

UDF_CHAR = "[a-zA-Z0-9_]"
UDF_FILE = "udf.sql"
PROCEDURE_FILE = "stored_procedure.sql"
ROUTINE_FILES = (UDF_FILE, PROCEDURE_FILE)
TEMP_UDF_RE = re.compile(f"(?:udf|assert)_{UDF_CHAR}+")
PERSISTENT_UDF_PREFIX_RE_STR = (
    r"CREATE\s+(?:OR\s+REPLACE\s+)?(?:FUNCTION|PROCEDURE)(?:\s+IF\s+NOT\s+EXISTS)?"
)
PERSISTENT_UDF_PREFIX = re.compile(PERSISTENT_UDF_PREFIX_RE_STR, re.IGNORECASE)
PERSISTENT_UDF_RE = re.compile(
    rf"{PERSISTENT_UDF_PREFIX_RE_STR}\s+(?:`?[a-zA-Z0-9_-]+`?\.)?`?({UDF_CHAR}+)`?\.`?({UDF_CHAR}+)`?",
    re.IGNORECASE,
)
UDF_NAME_RE = re.compile(r"^([a-zA-Z0-9_]+\.)?[a-zA-Z][a-zA-Z0-9_]{0,255}$")
GENERIC_DATASET = "_generic_dataset_"

raw_routines = {}


def get_routines_from_dir(project_dir):
    """Return all UDFs and stored procedures in the project directory."""
    return [
        {
            "name": root.split("/")[-2] + "." + root.split("/")[-1],
            "project": root.split("/")[-3],
            "is_udf": filename == UDF_FILE,
        }
        for root, dirs, files in os.walk(project_dir, followlinks=True)
        for filename in files
        if filename in ROUTINE_FILES
    ]


def get_routines(project):
    """Return all routines that could be referenced by the project."""
    return get_routines_from_dir(project) + get_routines_from_dir(
        Path(ConfigLoader.get("default", "sql_dir", fallback="sql")) / "mozfun"
    )


@attr.s(auto_attribs=True)
class RawRoutine:
    """Representation of the content of a single routine sql file."""

    filepath: str = attr.ib()
    name: str = attr.ib()
    dataset: str = attr.ib()
    project: str = attr.ib()
    definitions: List[str] = attr.ib([])
    tests: List[str] = attr.ib([])
    dependencies: List[str] = attr.ib([])
    description: str = attr.ib()
    is_stored_procedure: bool = attr.ib(False)

    @name.validator
    def validate_name(self, attribute, value):
        """Check that name is correctly derived."""
        filepath = Path(str(self.filepath))
        name = filepath.parent.name
        persistent_name = f"{self.dataset}.{name}"
        temp_name = f"{self.dataset}_{name}"

        if value is None or value != persistent_name:
            raise ValueError(
                f"Expected a function named {persistent_name} "
                f"or {temp_name} to be defined."
            )
        if not UDF_NAME_RE.match(name):
            raise ValueError(
                f"Invalid UDF name {name}: Must start with alpha char, "
                f"limited to chars {UDF_CHAR}, be at most 256 chars long"
            )

    @dataset.validator
    def validate_dataset(self, attribute, value):
        """Check that dataset name is valid."""
        if value != Path(self.filepath).parent.parent.name:
            raise ValueError("Invalid dataset name.")

    @project.validator
    def validate_project(self, attribute, value):
        """Check that project name is valid."""
        if value != Path(self.filepath).parent.parent.parent.name:
            raise ValueError("Invalid project name.")

    @dataset.default
    def set_default_dataset_name(self):
        """Set dataset default value to be derived from filepath."""
        return Path(str(self.filepath)).parent.parent.name

    @project.default
    def set_project_default_name(self):
        """Set project default value to be derived from filepath."""
        return Path(str(self.filepath)).parent.parent.parent.name

    @description.default
    def set_description_default(self):
        """Set description default value."""
        filepath = Path(str(self.filepath))
        metadata_file = filepath.parent / METADATA_FILE
        if metadata_file.exists():
            metadata = yaml.safe_load(metadata_file.read_text())
            if "description" in metadata:
                return metadata["description"]
        else:
            return ""

    @classmethod
    def from_file(cls, path):
        """Create a RawRoutine instance from text."""
        filepath = Path(path)
        text = render(
            filepath.name,
            template_folder=filepath.parent,
            format=False,
        )

        sql = sqlparse.format(text, strip_comments=True)
        statements = [s for s in sqlparse.split(sql) if s.strip()]

        name = filepath.parent.name
        dataset = filepath.parent.parent.name
        project = filepath.parent.parent.parent

        persistent_name_re = rf"`?{dataset}`?.`?{name}`?"
        persistent_name = f"{dataset}.{name}"
        temp_name = f"{dataset}_{name}"
        internal_name = None
        is_stored_procedure = False

        definitions = []
        tests = []

        procedure_start = -1

        for i, s in enumerate(statements):
            normalized_statement = " ".join(s.lower().split())
            if normalized_statement.startswith("create or replace function"):
                definitions.append(s)
                if re.search(persistent_name_re, normalized_statement):
                    internal_name = persistent_name

            elif normalized_statement.startswith("create temp function"):
                definitions.append(s)
                if temp_name in normalized_statement:
                    internal_name = temp_name

            elif normalized_statement.startswith("create or replace procedure"):
                is_stored_procedure = True
                definitions.append(s)
                tests.append(s)
                if re.search(persistent_name_re, normalized_statement):
                    internal_name = persistent_name

            else:
                if normalized_statement.startswith("begin"):
                    procedure_start = i

                if procedure_start == -1:
                    tests.append(s)

                if procedure_start > -1 and normalized_statement.endswith("end;"):
                    tests.append(" ".join(statements[procedure_start : i + 1]))
                    procedure_start = -1

        # get routines that could be referenced by the UDF
        routines = get_routines(project)
        dependencies = []
        for udf in routines:
            udf_re = re.compile(
                r"\b"
                + r"\.".join(f"`?{name}`?" for name in udf["name"].split("."))
                + r"\("
            )
            if udf_re.search("\n".join(definitions)):
                dependencies.append(udf["name"])

        dependencies.extend(re.findall(TEMP_UDF_RE, "\n".join(definitions)))
        dependencies = list(set(dependencies))

        if internal_name in dependencies:
            dependencies.remove(internal_name)

        return cls(
            name=internal_name,
            filepath=path,
            definitions=definitions,
            tests=tests,
            dependencies=sorted(dependencies),
            is_stored_procedure=is_stored_procedure,
        )


@attr.s(auto_attribs=True)
class ParsedRoutine(RawRoutine):
    """Parsed representation of a routine including dependent routine code."""

    tests_full_sql: List[str] = attr.ib([])

    @staticmethod
    def from_raw(raw_routine, tests_full_sql):
        """Promote a RawRoutine to a ParsedRoutine."""
        return ParsedRoutine(*attr.astuple(raw_routine), tests_full_sql)


def read_routine_dir(*project_dirs):
    """Read contents of routine dirs into dict of RawRoutine instances."""
    global raw_routines

    if not project_dirs:
        project_dirs = (ConfigLoader.get("default", "sql_dir"),)

    if project_dirs not in raw_routines:
        raw_routines[project_dirs] = {
            raw_routine.name: raw_routine
            for project_dir in project_dirs
            for root, dirs, files in os.walk(project_dir, followlinks=True)
            if os.path.basename(root) != ConfigLoader.get("routine", "example_dir")
            for filename in files
            if filename in ROUTINE_FILES
            for raw_routine in (RawRoutine.from_file(os.path.join(root, filename)),)
        }

    return raw_routines[project_dirs]


def parse_routines(project_dir):
    """Read routine contents of the project dir into ParsedRoutine instances."""
    # collect udfs to parse
    raw_routines = read_routine_dir(
        project_dir,
        Path(ConfigLoader.get("default", "sql_dir", fallback="sql")) / "mozfun",
    )

    # prepend udf definitions to tests
    for raw_routine in raw_routines.values():
        tests_full_sql = routine_tests_sql(raw_routine, raw_routines, project_dir)
        yield ParsedRoutine.from_raw(raw_routine, tests_full_sql)


def accumulate_dependencies(deps, raw_routines, udf_name):
    """
    Accumulate a list of dependent routine names.

    Given a dict of raw_routines and a udf_name string, recurse into the
    routine's dependencies, adding the names to deps in depth-first order.
    """
    if udf_name not in raw_routines:
        return deps

    raw_routine = raw_routines[udf_name]
    for dep in raw_routine.dependencies:
        deps = accumulate_dependencies(deps, raw_routines, dep)
    if udf_name in deps:
        return deps
    else:
        return deps + [udf_name]


def routine_usages_in_text(text, project):
    """Return a list of routine names used in the provided SQL text."""
    sql = sqlparse.format(text, strip_comments=True)
    routines = get_routines(project)

    udf_usages = []

    for routine in routines:
        if routine["name"] in sql:
            udf_usages.append(routine["name"])

    # the TEMP_UDF_RE matches udf_js, remove since it's not a valid UDF
    tmp_udfs = list(filter(lambda u: u != "udf_js", TEMP_UDF_RE.findall(sql)))
    udf_usages.extend(tmp_udfs)

    return sorted(set(udf_usages))


def routine_usage_definitions(text, project, raw_routines=None):
    """Return a list of definitions of routines used in provided SQL text."""
    if raw_routines is None:
        raw_routines = read_routine_dir()
    deps = []
    for udf_usage in routine_usages_in_text(text, project):
        deps = accumulate_dependencies(deps, raw_routines, udf_usage)
    return [
        statement
        for udf_name in deps
        for statement in raw_routines[udf_name].definitions
        if statement not in text
    ]


def sub_local_routines(test, project, raw_routines=None, stored_procedure_test=False):
    """
    Transform persistent UDFs into temporary UDFs for UDF tests.

    Use generic dataset for stored procedure tests.
    """
    if raw_routines is None:
        raw_routines = read_routine_dir()

    sql = prepend_routine_usage_definitions(test, project, raw_routines)

    for name, routine in raw_routines.items():
        if name in sql:
            for defn in routine.definitions:
                match = PERSISTENT_UDF_RE.match(defn)
                dataset, name = match.group(1), match.group(2)
                replace_name = f"{dataset}_{name}"
                if stored_procedure_test:
                    replace_name = f"{GENERIC_DATASET}.{replace_name}"
                sql = re.sub(
                    rf"(?<![\w\.])(`?{routine.project}`?\.)?`?{dataset}`?\.`?{name}`?(?=\()",
                    replace_name,
                    sql,
                )

    if not stored_procedure_test:
        sql = PERSISTENT_UDF_PREFIX.sub("CREATE TEMP FUNCTION", sql)
    return sql


def routine_tests_sql(raw_routine, raw_routines, project):
    """
    Create tests for testing persistent UDFs.

    Persistent UDFs need to be rewritten as temporary UDFs so that changes
    can be tested.
    """
    tests_full_sql = []
    for test in raw_routine.tests:
        test_sql = sub_local_routines(
            test, project, raw_routines, raw_routine.is_stored_procedure
        )
        tests_full_sql.append(test_sql)

    return tests_full_sql


def prepend_routine_usage_definitions(text, project, raw_routines=None):
    """Prepend definitions of UDFs used to provided SQL text."""
    statements = routine_usage_definitions(text, project, raw_routines)
    return "\n\n".join(statements + [text])



#!/usr/bin/env python3

"""Import application information from the probe info API."""

import json
import sys
from argparse import ArgumentParser

import requests
from google.cloud import bigquery

APP_LISTINGS_URL = "https://probeinfo.telemetry.mozilla.org/v2/glean/app-listings"

parser = ArgumentParser(description=__doc__)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="telemetry_derived")
parser.add_argument("--destination_table", default="applications_v1")
parser.add_argument("--sql_dir", default="sql/")
parser.add_argument("--dry_run", action="store_true")


def get_app_info():
    """Return a list of applications from the probeinfo API."""
    resp = requests.get(APP_LISTINGS_URL)
    resp.raise_for_status()
    return resp.json()


def main():
    """Run."""
    args = parser.parse_args()

    destination_table = (
        f"{args.project}.{args.destination_dataset}.{args.destination_table}"
    )

    bq_schema = (
        bigquery.SchemaField("app_channel", "STRING"),
        bigquery.SchemaField("app_description", "STRING"),
        bigquery.SchemaField("app_id", "STRING"),
        bigquery.SchemaField("app_name", "STRING"),
        bigquery.SchemaField("bq_dataset_family", "STRING"),
        bigquery.SchemaField("canonical_app_name", "STRING"),
        bigquery.SchemaField("document_namespace", "STRING"),
        bigquery.SchemaField("notification_emails", "STRING", mode="REPEATED"),
        bigquery.SchemaField("url", "STRING"),
        bigquery.SchemaField("v1_name", "STRING"),
        bigquery.SchemaField("is_glean", "BOOL"),
    )

    job_config = bigquery.LoadJobConfig(
        write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
    )
    job_config.schema = bq_schema

    app_info = get_app_info()
    app_data = []

    for data in app_info:
        app_data.append(
            {
                "app_channel": data.get("app_channel", "release"),
                "app_description": data.get("app_description"),
                "app_id": data.get("app_id"),
                "app_name": data.get("app_name"),
                "bq_dataset_family": data.get("bq_dataset_family"),
                "canonical_app_name": data.get("canonical_app_name"),
                "document_namespace": data.get("document_namespace"),
                "notification_emails": data.get("notification_emails"),
                "url": data.get("url"),
                "v1_name": data.get("v1_name"),
                "is_glean": True,
            }
        )

    # add legacy information
    legacy_desktop = {
        "app_channel": "release",
        "app_description": "Desktop version of Firefox using legacy telemetry",
        "app_id": None,
        "app_name": None,
        "bq_dataset_family": "telemetry",
        "canonical_app_name": "Firefox for Desktop",
        "document_namespace": "telemetry",
        "notification_emails": None,
        "url": "https://github.com/mozilla/gecko-dev",
        "v1_name": None,
        "is_glean": False,
    }

    for channel in ["nightly", "beta", "release"]:
        desktop_json = legacy_desktop.copy()
        desktop_json["app_channel"] = channel
        app_data.append(desktop_json)

    if args.dry_run:
        print(json.dumps(app_data))
        sys.exit(0)

    client = bigquery.Client(args.project)
    client.load_table_from_json(
        app_data, destination_table, job_config=job_config
    ).result()
    print(f"Loaded {len(app_data)} applications")


if __name__ == "__main__":
    main()



"""Publish UDFs and resources to the public mozfun GCP project."""

import fnmatch
import glob
import json
import os
import re
from argparse import ArgumentParser

from google.cloud import storage  # type: ignore
from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.routine.parse_routine import accumulate_dependencies, read_routine_dir
from bigquery_etl.util import standard_args
from bigquery_etl.util.common import project_dirs

OPTIONS_LIB_RE = re.compile(r'library = "gs://[^"]+/([^"]+)"')
OPTIONS_RE = re.compile(r"OPTIONS(\n|\s)*\(")


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--project-id",
    "--project_id",
    required=False,
    help="Project to publish UDFs to. "
    "If not set, publish UDFs for all projects except mozfun.",
)
parser.add_argument(
    "--target",
    default=ConfigLoader.get("default", "sql_dir", fallback="sql/")
    + ConfigLoader.get("default", "project"),
    required=False,
    help="Path to project directory.",
)
parser.add_argument(
    "--dependency-dir",
    "--dependency_dir",
    default=ConfigLoader.get("routine", "dependency_dir"),
    help="The directory JavaScript dependency files for UDFs are stored.",
)
parser.add_argument(
    "--gcs-bucket",
    "--gcs_bucket",
    default=ConfigLoader.get("routine", "publish", "gcs_bucket"),
    help="The GCS bucket where dependency files are uploaded to.",
)
parser.add_argument(
    "--gcs-path",
    "--gcs_path",
    default=ConfigLoader.get("routine", "publish", "gcs_path"),
    help="The GCS path in the bucket where dependency files are uploaded to.",
)
parser.add_argument(
    "--public",
    default=False,
    help="The published UDFs should be publicly accessible.",
)
standard_args.add_log_level(parser)
parser.add_argument(
    "pattern",
    default=None,
    nargs="?",
    help="glob pattern matching udfs to publish",
)


def main():
    """Publish routine."""
    args = parser.parse_args()

    if args.target is not None:
        projects = [args.target]
    else:
        projects = project_dirs()

    for project in projects:
        publish(
            args.target,
            args.project_id,
            os.path.join(
                ConfigLoader.get("default", "sql_dir"), project, args.dependency_dir
            ),
            args.gcs_bucket,
            args.gcs_path,
            args.public,
            pattern=args.pattern,
        )


def skipped_routines():
    """Get skipped routines from config."""
    return {
        file
        for skip in ConfigLoader.get("routine", "publish", "skip", fallback=[])
        for file in glob.glob(
            skip,
            recursive=True,
        )
    }


def publish(
    target,
    project_id,
    dependency_dir,
    gcs_bucket,
    gcs_path,
    public,
    pattern=None,
    dry_run=False,
):
    """Publish routines in the provided directory."""
    client = bigquery.Client(project_id)

    if dependency_dir and os.path.exists(dependency_dir):
        push_dependencies_to_gcs(
            gcs_bucket, gcs_path, dependency_dir, os.path.basename(target)
        )

    raw_routines = read_routine_dir(target)

    published_routines = []

    for raw_routine in (
        raw_routines if pattern is None else fnmatch.filter(raw_routines, pattern)
    ):
        # get all dependencies for UDF and publish as persistent UDF
        udfs_to_publish = accumulate_dependencies([], raw_routines, raw_routine)
        if pattern is not None:
            udfs_to_publish = fnmatch.filter(udfs_to_publish, pattern)
        udfs_to_publish.append(raw_routine)

        for dep in udfs_to_publish:
            if (
                dep not in published_routines
                and raw_routines[dep].filepath not in skipped_routines()
            ):
                publish_routine(
                    raw_routines[dep],
                    client,
                    project_id,
                    gcs_bucket,
                    gcs_path,
                    raw_routines.keys(),
                    public,
                    dry_run=dry_run,
                )
                published_routines.append(dep)


def publish_routine(
    raw_routine,
    client,
    project_id,
    gcs_bucket,
    gcs_path,
    known_udfs,
    is_public,
    dry_run=False,
):
    """Publish a specific routine to BigQuery."""
    if is_public:
        # create new dataset for routine if necessary
        dataset = client.create_dataset(raw_routine.dataset, exists_ok=True)

        # set permissions for dataset, public for everyone
        read_entry = bigquery.AccessEntry(
            "READER", bigquery.enums.EntityTypes.SPECIAL_GROUP, "allAuthenticatedUsers"
        )
        write_entry = bigquery.AccessEntry(
            "WRITER",
            bigquery.enums.EntityTypes.IAM_MEMBER,
            "group:team-data-platform@firefox.gcp.mozilla.com",
        )
        entries = list(dataset.access_entries)
        entries += [read_entry, write_entry]
        dataset.access_entries = entries
        dataset = client.update_dataset(dataset, ["access_entries"])

    # transforms temporary UDF to persistent UDFs and publishes them
    for definition in raw_routine.definitions:
        # Within a standard SQL function, references to other entities require
        # explicit project IDs
        for udf in set(known_udfs):
            # ensure UDF definitions are not replaced twice as would be the case for
            # `mozfun`.stats.mode_last and `mozfun`.stats.mode_last_retain_nulls
            # since one name is a substring of the other
            definition = definition.replace(f"`{project_id}.{udf}`", udf)
            definition = definition.replace(f"`{project_id}`.{udf}", udf)
            definition = definition.replace(f"{project_id}.{udf}", udf)
            definition = definition.replace(udf, f"`{project_id}`.{udf}")

        # adjust paths for dependencies stored in GCS
        query = OPTIONS_LIB_RE.sub(
            rf'library = "gs://{gcs_bucket}/{gcs_path}\1"', definition
        )

        # add UDF descriptions
        if (
            raw_routine.filepath not in skipped_routines()
            and not raw_routine.is_stored_procedure
        ):
            # descriptions need to be escaped since quotation marks and other
            # characters, such as \x01, will make the query invalid otherwise
            escaped_description = json.dumps(str(raw_routine.description))
            query = OPTIONS_RE.sub(f"OPTIONS(description={escaped_description},", query)

            if "OPTIONS(" not in query and query[-1] == ";":
                query = query[:-1] + f"OPTIONS(description={escaped_description});"

        print(f"Publish {raw_routine.name}")
        job = client.query(query, job_config=bigquery.QueryJobConfig(dry_run=dry_run))
        if not dry_run:
            job.result()


def push_dependencies_to_gcs(bucket, path, dependency_dir, project_id):
    """Upload UDF dependencies to a GCS bucket."""
    client = storage.Client(project_id)
    bucket = client.get_bucket(bucket)

    for root, dirs, files in os.walk(dependency_dir, followlinks=True):
        for filename in files:
            blob = bucket.blob(path + filename)
            blob.upload_from_filename(os.path.join(root, filename))


if __name__ == "__main__":
    main()



from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner
from google.api_core.exceptions import NotFound
from google.cloud.bigquery import SchemaField

from bigquery_etl.view import CREATE_VIEW_PATTERN, View

TEST_DIR = Path(__file__).parent.parent


class TestView:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    @pytest.fixture
    def simple_view(self):
        return View.from_file(
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "simple_view"
            / "view.sql"
        )

    @pytest.fixture
    def metadata_view(self):
        return View.from_file(
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "view_with_metadata"
            / "view.sql"
        )

    def test_from_file(self, simple_view):
        assert simple_view.dataset == "test"
        assert simple_view.project == "moz-fx-data-test-project"
        assert simple_view.name == "simple_view"
        assert (
            simple_view.view_identifier == "moz-fx-data-test-project.test.simple_view"
        )
        assert simple_view.is_user_facing

    def test_view_create(self, runner):
        with runner.isolated_filesystem():
            view = View.create("moz-fx-data-test-project", "test", "simple_view", "sql")
            assert view.path == Path(
                "sql/moz-fx-data-test-project/test/simple_view/view.sql"
            )
            assert "CREATE OR REPLACE VIEW" in view.content
            assert "`moz-fx-data-test-project.test.simple_view`" in view.content
            assert Path(
                "sql/moz-fx-data-test-project/test/dataset_metadata.yaml"
            ).exists()
            assert Path(
                "sql/moz-fx-data-test-project/test/simple_view/view.sql"
            ).exists()

            view = View.create(
                "moz-fx-data-test-project",
                "test",
                "simple_view",
                "sql",
                base_table="moz-fx-data-test-project.test.some_table",
            )
            assert "`moz-fx-data-test-project.test.some_table`" in view.content

    def test_view_invalid_path(self):
        with pytest.raises(ValueError):
            View(path="test", name="test", dataset="test", project="test")

    def test_view_valid(self, runner):
        with runner.isolated_filesystem():
            view = View.create("moz-fx-data-test-project", "test", "view", "sql")
            assert view.is_valid()

    def test_view_invalid(self, runner):
        with runner.isolated_filesystem():
            view = View.create("moz-fx-data-test-project", "test", "view", "sql")
            assert view.is_valid()

            view.path.write_text("CREATE OR REPLACE VIEW test.view AS SELECT 1")
            assert view.is_valid() is False

            view.path.write_text("SELECT 1")
            assert view.is_valid() is False

            view.path.write_text(
                "CREATE OR REPLACE VIEW `moz-fx-data-test-project.foo.bar` AS SELECT 1"
            )
            assert view.is_valid() is False

    def test_view_do_not_publish_invalid(self, runner):
        with runner.isolated_filesystem():
            view = View.create("moz-fx-data-test-project", "test", "view", "sql")
            assert view.is_valid()
            view.path.write_text("SELECT 1")
            assert view.is_valid() is False
            assert view.publish() is False

    @patch("google.cloud.bigquery.Client")
    @patch("google.cloud.bigquery.Table")
    def test_publish_valid_view(
        self, mock_bigquery_table, mock_bigquery_client, metadata_view
    ):
        mock_bigquery_client().get_table.return_value = mock_bigquery_table()

        assert metadata_view.is_valid()
        assert metadata_view.publish()
        assert mock_bigquery_client().update_table.call_count == 1
        assert (
            mock_bigquery_client().update_table.call_args[0][0].friendly_name
            == "Test metadata file"
        )
        assert (
            mock_bigquery_client().update_table.call_args[0][0].description
            == "Test description"
        )
        assert mock_bigquery_client().update_table.call_args[0][0].labels == {
            "123-432": "valid",
            "1232341234": "valid",
            "1234_abcd": "valid",
            "incremental": "",
            "incremental_export": "",
            "number_string": "1234abcde",
            "number_value": "1234234",
            "owner1": "test1",
            "owner2": "test2",
            "public_json": "",
            "schedule": "daily",
        }
        assert (
            mock_bigquery_client()
            .get_table("moz-fx-data-test-project.test.view_with_metadata")
            .friendly_name
            == "Test metadata file"
        )
        assert mock_bigquery_table().friendly_name == "Test metadata file"
        assert mock_bigquery_table().description == "Test description"

    @patch("bigquery_etl.dryrun.DryRun")
    @patch("google.cloud.bigquery.Client")
    def test_view_has_changes_no_changes(self, mock_client, mock_dryrun, simple_view):
        deployed_view = Mock()
        deployed_view.view_query = CREATE_VIEW_PATTERN.sub("", simple_view.content)
        deployed_view.schema = [SchemaField("a", "INT")]
        mock_client.return_value.get_table.return_value = deployed_view
        mock_dryrun.return_value.get_schema.return_value = {
            "fields": [{"name": "a", "type": "INT"}]
        }

        assert not simple_view.has_changes()

    def test_view_has_changes_non_matching_project(self, simple_view):
        assert not simple_view.has_changes(target_project="other-project")

    @patch("google.cloud.bigquery.Client")
    def test_view_has_changes_new_view(self, mock_client, simple_view, capsys):
        mock_client.return_value.get_table.side_effect = NotFound("")

        assert simple_view.has_changes()
        assert "does not exist" in capsys.readouterr().out

    @patch("google.cloud.bigquery.Client")
    def test_view_has_changes_changed_defn(self, mock_client, simple_view, capsys):
        deployed_view = Mock()
        deployed_view.view_query = """
        CREATE OR REPLACE VIEW
          `moz-fx-data-test-project.test.simple_view`
        AS
        SELECT
          99
        """
        mock_client.return_value.get_table.return_value = deployed_view

        assert simple_view.has_changes()
        assert "query" in capsys.readouterr().out

    @patch("google.cloud.bigquery.Client")
    def test_view_has_changes_changed_metadata(
        self, mock_client, metadata_view, capsys
    ):
        deployed_view = Mock()
        deployed_view.view_query = CREATE_VIEW_PATTERN.sub("", metadata_view.content)
        deployed_view.description = metadata_view.metadata.description
        deployed_view.friendly_name = metadata_view.metadata.friendly_name + "123"
        mock_client.return_value.get_table.return_value = deployed_view

        assert metadata_view.has_changes()
        assert "friendly_name" in capsys.readouterr().out

    @patch("bigquery_etl.dryrun.DryRun")
    @patch("google.cloud.bigquery.Client")
    def test_view_has_changes_changed_schema(
        self, mock_client, mock_dryrun, simple_view, capsys
    ):
        deployed_view = Mock()
        deployed_view.view_query = CREATE_VIEW_PATTERN.sub("", simple_view.content)
        deployed_view.schema = [SchemaField("a", "INT")]
        mock_client.return_value.get_table.return_value = deployed_view
        mock_dryrun.return_value.get_schema.return_value = {
            "fields": [{"name": "a", "type": "INT"}, {"name": "b", "type": "INT"}]
        }

        assert simple_view.has_changes()
        assert "schema" in capsys.readouterr().out
